(** * Theory of Sets : Miscellaneous results
  Copyright INRIA (2012-2013) Marelle Team (Jose Grimm). 
*)


Require Import ssreflect ssrbool eqtype ssrfun.
(* $Id: ssete11.v,v 1.5 2016/05/19 08:34:58 grimm Exp $ *)
Set Implicit Arguments.
Unset Strict Implicit.
Import Prenex Implicits.

Require Export sset14.

Module AFA.

(** ** Anti foundation axiom
 We study some properties that are forbidden by the axiom of foundation *)

Lemma afa_ex1 x: x <> powerset x.
Proof.
set A := Zo x (fun z => ~ inc z z) => h.
have aux: ~(inc A A).
  move /Zo_P => [pa pb]; apply: (pb); apply/Zo_P; split => //.
apply: (aux); apply:Zo_i => //; rewrite h; apply /setP_P; apply:Zo_S.
Qed.

Lemma afa_ex2 a b: a = singleton b -> b = doubleton emptyset a ->
  a = singleton (powerset a).
Proof. by move => pa pb; rewrite {2} pa setP_1 - pa - pb. Qed.

Lemma afa_ex2_inv a: a = singleton (powerset a) ->
  exists b,  a = singleton b /\ b = doubleton emptyset a.
Proof.
set b := powerset a => eq1; exists b; split => //.
by move: (setP_1 b); rewrite  - eq1.
Qed.

Lemma afa_ex3 a b c d: a = singleton b -> 
  b = (doubleton emptyset (singleton emptyset)) +s1 c +s1 d ->
  c = doubleton emptyset a -> d = singleton a ->
  a = singleton (powerset (powerset a)).
Proof. 
move => av bv cv dv; rewrite av setP_1 {1} bv -av; congr (singleton _).
set_extens t.
  move => h; apply/setP_P; case /setU1_P :h.
    case /setU1_P.
      case /set2_P => ->; fprops;apply : set1_sub; fprops.
    move => ->; rewrite - cv; fprops.
  move => ->; rewrite dv; apply : set1_sub; fprops.
move /setP_P => h.
case: (inc_or_not emptyset t) => ta; case: (inc_or_not a t) => tb.
- apply /setU1_P; left; apply /setU1_P; right; rewrite cv.
  by apply: extensionality => // => s; case /set2_P => ->.
- apply /setU1_P; left; apply /setU1_P; left; apply /set2_P; right.
  apply set1_pr => // s st; move: (h _ st); case /set2_P => // sa; case: tb; ue.
- apply /setU1_P; right; rewrite dv; apply set1_pr => // s st.
  move: (h _ st); case /set2_P => // sa; case: ta; ue.
- apply /setU1_P; left; apply /setU1_P; left; apply /set2_P; left.
  apply /set0_P => // s st; move: (h _ st); case /set2_P => sa.
    case: ta; ue.
  case: tb; ue.
Qed.


Lemma afa_ex3_inv a : a = singleton (powerset (powerset a)) ->
  exists b c d, [/\ 
    a = singleton b,
    b = (doubleton emptyset (singleton emptyset)) +s1 c +s1 d,
    c = doubleton emptyset a &
    d = singleton a].
Proof. 
set b := (powerset (powerset a)) => av; exists b.
move: (setP_1 b); rewrite - av; set c := doubleton emptyset a => h.
set d:= singleton a;exists c, d; split => //.
set_extens t.
  move => /setP_P; rewrite av; rewrite setP_1 - av => ct.
  case: (inc_or_not emptyset t);case: (inc_or_not a t) => ha hb.
  - apply: setU1_r; apply/setU1_P; right; set_extens u; first by move/ct. 
    by case/set2_P => ->.
  - apply: setU1_r;apply: setU1_r; apply/set2_P; right.
    by apply:(set1_pr hb) => u ut; case/set2_P:(ct _ ut) => // ua; case: ha;ue.
  - apply/setU1_P; right; apply:(set1_pr ha) => u ut.
   case/set2_P:(ct _ ut) => // ua; case: hb;ue.
  - apply: setU1_r;apply: setU1_r; apply/set2_P; left; apply/set0_P => u ut.
   case/set2_P:(ct _ ut) => hc; [ case: hb; ue | case: ha;ue].
case/setU1_P; last by move ->; apply /setP_P => u /set1_P ->;  apply: setP_Ti.
case/setU1_P; last by move ->;  rewrite - h; apply: setP_Ti.
case /set2_P => ->; first by apply: setP_0i.
by apply /setP_P => u /set1_P ->; apply:setP_0i.
Qed.

Lemma afa_ex4 x: x = gfunctions emptyset x <-> x = singleton emptyset.
Proof. by rewrite gfunctions_empty. Qed.

Lemma afa_ex5 X A: X = gfunctions X A <-> 
  (exists a f, [/\ A = singleton a, X = singleton f & f = singleton (J f a)]).
Proof.
split; last first.
  move => [a [f [Aa Xf fv]]].
  symmetry; rewrite {2} Xf.
  apply set1_pr.
    move: (simple_fct2 f a) => [sa sb _ _ _].
    apply /gfunctions_P1; rewrite fv sb; split => //.  
    move => t /set1_P; rewrite  Aa Xf => ->; apply: setXp_i; fprops.
  move => e /gfunctions_P1 [pa pb pc]; rewrite fv.
  apply: set1_pr1.
    apply /domain_set0P; rewrite pb Xf; fprops.
  move => z ze; move: (pc _ ze); rewrite Aa Xf. 
  by move /setX_P => [qa ] /set1_P qb/set1_P qc; rewrite - qa qb qc.
move => xsg.
case: (emptyset_dichot X) => xne.
  by empty_tac1 emptyset; rewrite {2} xsg xne gfunctions_empty; apply:set1_1.
move: xne => [f fX].
move: (fX); rewrite xsg; move => /gfunctions_P2 [pa pb pc].
set a := Vg f f.
have aA: inc a A.
  apply: pc; apply /range_gP => //; rewrite pb; ex_tac.
case: (equal_or_not A (singleton a)) => Aa.
  have pd: gfunctions X A = singleton f.
    apply: set1_pr; first by ue.
    move => z /gfunctions_P2 [qa qb qc].
    apply: fgraph_exten => //; first by ue.
    move => t td /=.
    have: inc (Vg z t) A by apply qc;  apply /(range_gP qa); ex_tac. 
    have: inc (Vg f t) A. 
       by apply pc; apply /(range_gP pa); rewrite pb - qb; ex_tac.  
    by rewrite Aa; move => /set1_P -> /set1_P ->.
  exists a; exists f;split => //.
  apply: set1_pr; first by apply: fdomain_pr1 => //; rewrite pb.
  move => e zf; rewrite (in_graph_V pa zf).
  by move: (domain_i1 zf); rewrite pb xsg pd => /set1_P ->.
have [b bA ba]: exists2 b, inc b A & b <> a. 
  ex_middle bad; case: Aa; apply: set1_pr => // z zA.
  ex_middle h; case: bad; ex_tac.
set F := Lg X (fun z => Yo (Vg z z = a) b a).
have fgf: fgraph F by rewrite /F; fprops.
have df: domain F = X by rewrite /F; bw.
have fx: inc F X. 
  rewrite xsg; apply /gfunctions_P2;split => //.
  by move => t /(range_gP fgf) [x qa ->]; rewrite /F; bw; try ue; Ytac h.
case: (equal_or_not (Vg F F) a) => eq1.
   move: (eq1); rewrite {1} /F; bw; Ytac0 => eba; by case: ba.
by move: (eq1); rewrite {1} /F; bw; Ytac0; case.
Qed.

Lemma afa_ex6 X: X = gfunctions X X <-> 
  (exists2 x, X = singleton x & x = singleton (J x x)).
Proof.
split.
  move/afa_ex5 => [a [f [pa pb pc]]]; exists a => //.
  by rewrite pa in pb; rewrite {1 2} (set1_inj pb).
by move => [x pa pb]; apply /afa_ex5; exists x; exists x.
Qed.

Lemma afa_ex7 X: X = gfunctions X X <-> 
   X = singleton (singleton (singleton X)).
Proof.
apply (iff_trans (afa_ex6 X)).
have pa: forall t,  J t t = singleton (singleton t).
  by  move => t;  rewrite Pair.kprE /kpair /singleton.
split.
  by move => [a ab]; rewrite pa => s; rewrite ab - s.
move => h; exists (singleton (singleton X)) => //. 
by rewrite pa - h.
Qed.

Lemma afa_ex8 X: X = singleton X -> X = gfunctions X X. 
Proof. by move => h; apply /afa_ex7; rewrite -h -h. Qed.

Lemma afa_ex9 X A: X = functions X A <-> 
  (exists a f, [/\ A = singleton a, X = singleton f & f = Lf (fun _ => a) X A]).
Proof.
have res: forall A a f t,
     function f ->  A = singleton a -> target f = A -> 
     inc t (source f) -> Vf f t = a.
  by move => B a f t ff -> tf tsf;move:(Vf_target ff tsf);rewrite tf => /set1_P.
split; last first. 
  move => [a [f [Aa Xf fv]]].
  symmetry; rewrite {2} Xf.
  have la: lf_axiom (fun _ : Set => a) X A by move =>  t _; rewrite Aa; fprops.
  have aux: function (Lf (fun _ : Set => a) X A) by apply: lf_function.
  apply: set1_pr.
    apply /functionsP; rewrite fv;red;aw;split => //.
  move => e /functionsP [pa pb pc]; rewrite fv. 
  apply function_exten => //; aw => z ze; aw; last by ue. 
  by apply: (@res A). 
move => xsg.
case: (emptyset_dichot X) => xne.
  have aux:= (empty_function_tg_function A).
  by empty_tac1 ( empty_function_tg A); rewrite xsg xne; apply /functionsP.
move: xne => [f fX].
move: (fX); rewrite xsg; move => /functionsP [pa pb pc].
set a := Vf f f.
have aA: inc a A by rewrite -pc /a; Wtac.
case: (equal_or_not A (singleton a)) => Aa.
  have ha: functions X A = singleton f.
    apply: set1_pr;first by ue.
    move => z /functionsP [qa qb qc].
    apply: function_exten => //; (try ue); move => t td /=.
    have tx: inc t (source f) by rewrite pb - qb.
    by rewrite (res A a f t pa Aa pc tx) (res A a z t qa Aa qc td).  
  have pf: lf_axiom (fun _ => a) (functions X A) A by move => t _.
  have sf: source f = (functions X A) by ue.
  exists a, f;split => //.
  apply: function_exten; aw. 
  - by apply: lf_function.
  - move => t tf /=; aw; [ by apply: (@res A) | ue ].
have [b bA ba]: exists2 b, inc b A & b <> a. 
  ex_middle bad; case: Aa; apply: set1_pr => // z zA.
  ex_middle h; case: bad; ex_tac.
set lam := (fun z => Yo (Vf z z = a) b a).
have ta: lf_axiom lam X A by rewrite /lam; move => t tX /=; Ytac h.
have fx: inc (Lf lam X A) X.
  rewrite {2} xsg; apply /functionsP;red;aw;split => //.
  by apply: lf_function.  
by case: (equal_or_not (Vf (Lf lam X A)(Lf lam X A)) a) => eq1;
  move: (eq1); aw; rewrite {1} /lam; Ytac0.
Qed.

Lemma afa_ex10 X : X = functions X X <-> 
  (exists2 f, X = singleton f & f = triple X X  (singleton (J f f))).
Proof.
have aux: forall f, X = singleton f  -> 
  triple X X (singleton (J f f)) = Lf (fun _ : Set => f) X X.
  move => f Xf;move: (simple_fct2 f f); set g := (singleton (J f f)).
  move => [fgg sb sc sd se].
  have pa: lf_axiom (fun _ => f) X X by move => s _ ; rewrite Xf;apply: set1_1.
  apply function_exten; aw.
  + by rewrite Xf;apply: function_pr => //; rewrite sc.
  + by apply: lf_function. 
  + move => t tx /=; rewrite {1} /Vf; aw.
    by move: tx; rewrite  Xf  => /set1_P => ->. 
apply: (iff_trans (afa_ex9 X X)); split.
  move => [a [f [xa xb xc]]]; exists f => //; move: xc.
  rewrite xb in xa; rewrite -(set1_inj xa) => {1} ->.
  by symmetry; apply: aux.
by move => [f Xf tf]; exists f, f;split => //; rewrite {1} tf; apply: aux.
Qed.

Lemma afa_ex11 X : X = functions X X <-> 
  X = singleton (singleton (singleton (singleton (singleton X)))).
Proof.
split.
  move /afa_ex10 => [f ->].
  by rewrite /triple Pair.kprE /kpair - !/(singleton _) => <-.
move => h; apply /afa_ex10.
exists (singleton (singleton (singleton (singleton X)))); first exact.
by rewrite /triple Pair.kprE /kpair -! /(singleton _) -h  -! /(singleton _).
Qed.

Lemma afa_ex12 X : X = singleton X ->  X = functions X X.
Proof. by move => h; apply /afa_ex11; rewrite -!h. Qed.

End AFA.

(** * Lemmas that do not use cardinal or ordinal numbers *)
Module Misc1.

(** There is a distributive law between [powerset] and [intersection2] *)

Lemma powerset_mono A B: sub A B -> sub (powerset A)(powerset B).
Proof.
move=> sAB t /setP_P ta; apply/setP_P; apply:(sub_trans ta sAB).
Qed.

Lemma union_monotone3 A B:
  sub A B -> sub (union A) (union B).
Proof. 
move=> sAB t /setU_P [y ty yA]; apply/setU_P;exists y => //; fprops.
Qed.

Lemma intersection_greatest A B x: 
  sub x A -> sub x B -> sub x (A \cap B).
Proof. move=> xA xB t tx; apply /setI2_P; split;fprops. Qed.

Lemma powerset_inter: {morph powerset : A B / A \cap B}.
Proof.
move => A B; apply: extensionality.
  apply: intersection_greatest; apply: powerset_mono. 
     apply: subsetI2l.
  apply:subsetI2r.
move=> x /setI2_P[] /setP_P xpA /setP_P xpB; apply /setP_P. 
by apply: intersection_greatest.
Qed.


(** The following gives a well-ordering on [doubleton a b], even when [a=b] *)

Definition example_worder a b := (tripleton (J a a) (J b b)  (J a b)).

Lemma example_worder_gleP a b x y: 
    related  (example_worder a b) x y <->
    [\/ (x = a /\ y = a),  (x = b /\ y = b) |  (x = a /\ y = b)].
Proof.
split; first by case /set3_P => h;rewrite(pr1_def h)(pr2_def h); in_TP4.
case => [] [->  ->]; apply /set3_P; in_TP4.
Qed.

Lemma example_is_worder a b:
  worder_on (example_worder a b) (doubleton a b).
Proof. 
have gc:sgraph (example_worder a b).
  move => t /set3_P; case => ->; fprops.
have subs: substrate (example_worder a b) = doubleton a b.
  set_extens x. 
    case/(substrate_P gc)=> [] [y].
      move / (example_worder_gleP a b x y); case; case => ->; fprops.
    move /(example_worder_gleP a b y x); case; case => _ ->; fprops.
  case /set2_P=> h; have aux: (related (example_worder a b) x x)
    by apply/(example_worder_gleP a b x x); in_TP4.
    substr_tac.
  substr_tac.
have pa: related (example_worder a b) a a by  apply/example_worder_gleP; in_TP4.
have pb: related (example_worder a b) b b by  apply/example_worder_gleP; in_TP4.
have pc: related (example_worder a b) a b by  apply/example_worder_gleP; in_TP4.
have oc: (order (example_worder a b)). 
  split => //.
    by red;rewrite subs => y /set2_P [] ->. 
  move=> x y z/example_worder_gleP; case; move => [ -> ->]//.
    move /example_worder_gleP; case; move => [ h ->] //.
  move=> x y /example_worder_gleP; case;move => [ -> ->] //.
  by move /example_worder_gleP; case; case.
split; [split; [  exact | ] | exact ].
move => x xab nex.
case: (inc_or_not a x)=> hyp.
  exists a; red; rewrite (iorder_sr oc xab); split=>//.  
    move=> y yx; apply /(iorder_gleP _ hyp yx); apply /example_worder_gleP.
    move: (xab _  yx); rewrite subs; case /set2_P; in_TP4.
move: nex=> [y yx]; exists y; red;rewrite (iorder_sr oc xab); split=>//.
move => z zx; apply /(iorder_gleP _ yx zx); apply /example_worder_gleP.
rewrite subs in xab.
case /set2_P: (xab _  yx) => yb; first by case: hyp; rewrite -yb.
case /set2_P: (xab _  zx) => zb; first by case: hyp; rewrite - zb.
in_TP4.
Qed.


(** Example of inductive sets (another one is in the main text)  *)

Lemma inductive_example1 A F: sub A (powerset F) ->
  (forall S, (forall x y, inc x S -> inc y S -> sub x y \/ sub y x) ->
    inc (union S) A) ->
  inductive (sub_order A).
Proof. 
move=> Ap ih X Xs [oX tX];exists (union X).
have  [oi si ] := (sub_osr A).
move: tX; aw => tX;rewrite si in Xs.
have uA: inc (union X) A.
  by apply: ih; move=> x y xX yX; move: (tX _ _ xX yX); case =>h;
   move: (iorder_gle1 h) => /sub_gleP [_ _ h1]; [left | right].
split; first by rewrite si.
by move=> y yX; apply /sub_gleP;split; [apply: Xs | exact | apply: setU_s1].
Qed.


End Misc1.

Module Zermelo.


(** Direct proof of the transfinite principle  *)

Lemma transfinite_principle_bis r (p:property):
  worder r ->  
  (forall x, inc x (substrate r) ->
    (forall y, glt r y x -> p y) -> p x) -> 
  forall x, inc x (substrate r) -> p x.
Proof. 
move => [or wor] hyp x xsr; ex_middle npx.
set (X:=Zo (substrate r) (fun x => ~ p x)). 
have neX: (nonempty X) by exists x; apply: Zo_i.
have Xsr: sub X (substrate r) by apply: Zo_S.  
move:(wor _ Xsr neX)=> [y []]; aw => /Zo_P [ysr npy] yle.
case: npy; apply: hyp =>//.
move=> t ty; ex_middle npt.
have tsr: inc t (substrate r) by order_tac.
move: (iorder_gle1 (yle _ (Zo_i tsr npt))) => nty; order_tac.
Qed.

Definition segment_rp r s:=
  sub s (substrate r) /\ (forall x y, inc x s -> gle r x y -> inc y s).

Lemma segment_rpC r x : order r -> sub x (substrate r) ->  
    (segmentp r x <-> segment_rp r (substrate r -s x)).
Proof.
move => or xsr; split => [][sa sb].
  split; first by apply: sub_setC.
  move => a b /setC_P [asr nax] lab; apply:setC_i; first by order_tac.
  dneg bx; apply: (sb _ _ bx lab).
split => // a b ax lba; ex_middle bnx.
have iby: inc b (substrate r -s x) by apply/setC_P; split => //; order_tac.
by case/setC_P: (sb _ _ iby lba). 
Qed.

Lemma segment_rpC' r x : order r -> sub x (substrate r) ->  
    (segment_rp r x <-> segmentp r (substrate r -s x)).
Proof.
move => or xsr; apply:iff_sym.
rewrite - {2}(setC_K xsr). apply:(segment_rpC or); apply:(sub_setC). 
Qed.

Lemma segment_rC r x: total_order r -> inc x (substrate r) ->
  (segment_r r x) = substrate r -s (segment r x).
Proof. 
move => [or tor] xsr; set_extens y.
  move => /segment_rP ha; apply /setC_P; split; first by order_tac.
  move/segmentP => hb; order_tac. 
move => /setC_P [ysr /segmentP ha]; apply/segment_rP.
case: (equal_or_not y x) => exy; first by rewrite exy; order_tac.
by case: (tor _ _ xsr ysr) => // hb; case: ha.
Qed.

Lemma segment_rC' r x: total_order r -> inc x (substrate r) ->
  (segment r x) = substrate r -s (segment_r r x).
Proof.
by move=> sa sb; rewrite (segment_rC sa sb) setC_K //; apply: sub_segment.
Qed.

Lemma segment_r_p0 r: segment_rp r emptyset. 
Proof. by split; fprops => cx y /in_set0. Qed.

Lemma segment_r_p1 r x: order r -> inc x (substrate r) -> 
  segment_rp r (segment_r r x). 
Proof.
move => or xsr.
split; first by move => y /segment_rP le; order_tac.
move => a b /segment_rP le1 le2; apply/segment_rP; order_tac.
Qed.

Lemma segment_rp_p2 r z:
  order r -> inc z (substrate r) -> segment_rp r (segment_r r z -s1 z).
Proof.
move => or zr.
split; first by move => t /setC1_P [/segment_rP le _]; order_tac.
move => x y /setC1_P [/segment_rP le1 ne1] le2; apply/setC1_P; split.
  apply/segment_rP; order_tac.
by dneg xx; rewrite xx in le2; order_tac.
Qed.

Lemma segment_rp_p3 r s: order r ->  
   (alls s (segment_rp r)) -> segment_rp r (intersection s).
Proof. 
move => or al.
case: (emptyset_dichot s) => nes. 
  rewrite nes setI_0; apply: segment_r_p0.
move:(nes) => [k ks].
split.
  move => t ti; exact: (proj1 (al _ ks) _ (setI_hi ti ks)).
move => x y xI lxy; apply: (setI_i nes)  => a ias.
move: (al _ ias) => [_ sy]; exact :(sy _ _ (setI_hi xI ias) lxy).
Qed.


Lemma segment_r_succ_or_limit r x: worder r -> inc x (substrate r) ->
  [\/ (segment_r r x = substrate r),
  (exists2 z, inc z (segment r x) & segment_r r x = segment_r r z -s1 z) |
  (segment_r r x) = intersection (fun_image (segment r x) (segment_r r)) ].
Proof.
move => wor xsr.
have tor:= (worder_total wor).
have or := proj1 wor.
set T := segment r x.
case: (emptyset_dichot T) => te.
  by move: (segment_rC tor xsr); rewrite -/T te setC_0 => h; constructor 1.
have TE: sub T (substrate r) by apply: sub_segment.
have Tb: bounded_above r T by exists x; split => // y /segmentP [].
move: (worder_hassup wor TE Tb) => [z /(lubP or TE) [[za zb] zc]].
case: (inc_or_not z T) => zT.
  constructor 2; ex_tac.
  have ltx: glt r z x by apply /segmentP.
  set_extens u.
     move /segment_rP => l1.
     have [l2 /nesym n2]: glt r z u by order_tac.
     by apply/setC1_P; split => //;apply /segment_rP.
  move => /setC1_P [/segment_rP qa qb]; apply /segment_rP.
  have uE: inc u (substrate r) by order_tac.
  case: (equal_or_not u x) => eq; first by rewrite eq; order_tac. 
  case: (proj2 tor _ _ xsr uE) => // le1.
  move /segmentP: (conj le1 eq) => /zb  le2; case: qb; order_tac.
constructor 3.
have nf: nonempty (fun_image T (segment_r r)) by apply:funI_setne.
set_extens t.
  move /segment_rP => h; apply /(setI_P nf) => i /funI_P [u /segmentP zu ->].
  move: zu => [zu _];apply /segment_rP; order_tac.
move => /(setI_P nf) h; apply /segment_rP.
have hh: forall q, inc q T -> gle r q t.
  move => q qt;apply /segment_rP; apply: h; apply/funI_P; ex_tac.
have tE: inc t (substrate r) by move: te => [u /hh ut];  order_tac. 
have zt : gle r z t by apply:zc; split => //; apply: hh.
case: (equal_or_not z x) => eq; first by rewrite - eq.
have zE: inc z (substrate r) by order_tac.
case: (proj2 tor _ _ xsr zE) => // le1;first order_tac. 
by case: zT; apply /segmentP.
Qed.

Definition segment_rs r:= (fun_image (substrate r) (segment_r r) +s1 emptyset).

Lemma segment_rs_P r x: 
  (inc x (segment_rs r) <-> 
     x = emptyset \/ (exists2 y, inc y (substrate r) & x = segment_r r y)).
Proof.
split; first by case /setU1_P; [ by move/funI_P; right | by left].
by move => ww; apply/setU1_P; case:ww; [right| left; apply/funI_P].
Qed.

Lemma segment_rs_p1 r: worder r ->
  forall x, (inc x (segment_rs r) <-> segment_rp r x).
Proof.
move => wor x; apply:(iff_trans (segment_rs_P r x)); split.
  case; first by  move => ->;  apply:segment_r_p0.
  move => [y ysr ->]; exact:(segment_r_p1 (proj1 wor) ysr).
move => ss.
move /(segment_rpC' (proj1 wor) (proj1 ss)): (ss) => ss'.
case:(well_ordered_segment wor ss').
  by move => sa; left; rewrite -(setC_K (proj1 ss)) sa setC_v.
move => [y ysr yv]; right; exists y => //.
by rewrite (segment_rC (worder_total wor) ysr) - yv (setC_K (proj1 ss)).
Qed.


Lemma Zermelo_omega_chain r (E := substrate r): Zermelo_like r -> 
  Zermelo_chain E (segment_rs r).
Proof.
move => [ha hb]; move: (ha) => [or wor].
have ra: sub (segment_rs r) (powerset E).
  by move => t /(segment_rs_p1 ha) [/setP_P ts _].
have rb: inc (substrate r) (segment_rs r).
  apply/(segment_rs_p1 ha); split => // x y xsr le; order_tac.
have rc: forall A, inc A (segment_rs r)  -> inc (A -s1 rep A) (segment_rs r).
  move => A /segment_rs_P; case.
    by move => ->; apply /setU1_P; right;apply/set0_P => t /setC1_P [/in_set0].
  move => [z zE ->]; rewrite (hb _ zE). 
  by apply/(segment_rs_p1 ha) /segment_rp_p2.
split => // A Aom neA; apply/(segment_rs_p1 ha). 
by apply:(segment_rp_p3 or) => t /Aom /(segment_rs_p1 ha).
Qed.

Lemma Zermelo_chain_least E F: Zermelo_chain E F -> inc emptyset F.
Proof.
move => [pa pb pc pd].
have nef: nonempty F by exists E; apply: pb.
move: (pd _ (@sub_refl F) nef); set x := intersection F => iF.
case: (emptyset_dichot x); [by move <- | move => xne].
move: (pc _ iF) => siF.
by move: (setI_hi  (rep_i xne) siF) => /setC1_P [].
Qed.

Lemma Zermelo_chain_minimal1 r (E := substrate r):
  Zermelo_like r -> 
  forall F, Zermelo_chain E F -> sub (segment_rs r) F.
Proof.
move => [wor ha] F zl => t /(segment_rs_P).
have eF := (Zermelo_chain_least zl).
case; first by move => ->.
have or := proj1 wor.
move: (zl) => [pa pb pc pd] [k kE ->].
move: k kE. apply:(transfinite_principle_bis wor) => x xsr H.
case: (segment_r_succ_or_limit wor xsr).
- by move ->.
- move => [z /segmentP ltx ->]; move: (pc _ (H _ ltx)); rewrite ha //.
  order_tac.
- move ->; case: (emptyset_dichot (segment r x)) => te.
    by rewrite te funI_set0 setI_0.
  apply:pd; last by apply: funI_setne.
  by move => u /funI_P [z /segmentP /H h ->].
Qed.

Lemma Zermelo_chain_minimal r (E := substrate r):
  Zermelo_like r -> 
  (segment_rs r) = intersection (Zo (powerset (powerset E)) (Zermelo_chain E)).
Proof.
move => h.
set Z := Zo _ _.
have w: forall F, inc F Z <-> Zermelo_chain E F.
  move => f; split;first by move/Zo_hi.
  by move => zc; apply: Zo_i => //; apply /setP_P; move: zc=> [].
have ha: inc (segment_rs r) Z by apply/w/Zermelo_omega_chain.
set_extens t; last by move => ts; exact:(setI_hi ts ha).
move => ts; apply/setI_P; [ex_tac | move => i /w ci].
exact: (Zermelo_chain_minimal1 h ci ts).
Qed.

Lemma Zermelo_unique r r': Zermelo_like r -> Zermelo_like r' ->
   substrate r = substrate r' -> r = r'.
Proof. 
move => ha hb ss.
have ch12: (segment_rs r) =  (segment_rs r'). 
  rewrite (Zermelo_chain_minimal ha).
  by rewrite (Zermelo_chain_minimal hb) ss.
move: ha hb => [wor1 ha] [wor2 hb].
move:(wor1)(wor2) => [or1 _][or2 _].
have pa: forall x, inc x (substrate r) -> segment_r r x = segment_r r' x.
  move => x xsr.
  have: inc (segment_r r x) (segment_rs r'). 
     by rewrite - ch12; apply/setU1_P; left; apply/funI_P; ex_tac.
  case /setU1_P.
    move/funI_P => [z za ea]; rewrite ea.
    by move:(f_equal rep ea); rewrite (ha _ xsr) (hb _ za) => ->.
  by move => h; empty_tac1 x; apply/segment_rP; order_tac.
apply:order_exten => // x y.
case: (inc_or_not x (substrate r)) => xE; last first.
  split => h; case xE; [ | rewrite ss ]; order_tac.
have ssx:=(pa  _ xE).
split => /segment_rP; first by rewrite ssx => /segment_rP.
by rewrite - ssx => /segment_rP.
Qed.

Lemma Zermelo_like_chain_least1 r X:
  Zermelo_like r ->  sub X (substrate r) -> nonempty X ->
  exists!Y, [/\ sub X Y, inc Y (segment_rs r) & inc (rep Y) X].
Proof.
move => [pa pb] XE neX.
have or: order r by fprops.
apply/unique_existence; split.
  move:(proj2 pa  _ XE neX) => [y []].
  rewrite (iorder_sr or XE)=> yX yP.
  have yE := (XE _ yX). 
  have ha:sub X (segment_r r y). 
     by move => t /yP /iorder_gle1 tX; apply /segment_rP.
  move/(segment_rs_p1 pa): (segment_r_p1 or yE) => hb.
  have hc:inc (rep (segment_r r y)) X by rewrite pb.
  by exists (segment_r r y).
move => Y1 Y2 [qa qb qc] [qd qe qf].
have [x0 x0X] := neX.
case/segment_rs_P: qb; last move => [y1 y1sr Y1v].
   by move => ye; move:(qa _ x0X); rewrite ye => /in_set0.
case /segment_rs_P:qe; last move =>  [y2 y2sr Y2v].
   by move => ye; move:(qd _ x0X); rewrite ye => /in_set0.
move: qc qf qa qd; rewrite Y1v Y2v pb // pb // => qc qf y1X y2X.
move:(y2X _ qc); move/segment_rP => sa.
move:(y1X _ qf); move/segment_rP => sb.
f_equal; order_tac.
Qed.

(* bof 
Lemma Zermelo_w  E W: sub W (powerset E) -> (inc emptyset W) ->
  (forall X,  sub X E -> nonempty X ->
  exists!Y, (sub X Y /\ inc Y W /\ inc (rep Y) X)) ->
  Zermelo_chain E W.
Proof.
move => wp eW H.
have EW: inc E W. 
  case: (emptyset_dichot E) => neE; first by rewrite neE.
   move: (H E (@sub_refl E) neE) => [x [ [pa [pb _]] _]].
   move: (wp _ pb) => /setP_P xE.
   by rewrite(extensionality pa xE).
have Ha: forall A B, inc B W -> sub A B -> inc (rep B) A -> inc A W -> A = B.
  move => A B BW saB rBA AW.
  have sae: sub A E by move/setP_P:(wp _ BW); apply: (sub_trans).
  have neA: nonempty A by exists (rep B).
  move:(rep_i neA) (@sub_refl A) => raa saa.
  by move: (H A sae neA) => /unique_existence [_]; apply.
Abort.
*)

End Zermelo.

(** ** The von Neuman ordinals *)
Module VonNeumann. 


(* Definition of von Neumann of an ordinal *)
Lemma von_neumann_ordinal E:
  ordinalp E <->    
    (worder (ordinal_o E) /\
    (forall x, inc x E -> segment (ordinal_o E) x = x)).
Proof.
split.
   move => oE; split; first by apply: (ordinal_o_wor oE).
   by move => x xE;apply:ordinal_segment.
move => [sa sb].
have Ha y x: inc x E -> inc y x -> ssub y x.
  by move => xE; rewrite -{1} (sb _ xE); move/segmentP => [/ordo_leP []]. 
have tE: transitive_set E.
  by move => x xE y; rewrite - (sb _ xE); move/segmentP => [/ordo_leP []].
have aE: asymmetric_set E. 
  move => y z yE zE /(Ha _ _ zE) [pa pb] /(Ha _ _ yE) [pc _].
  by case: pb; apply:extensionality.
suff: (ordinal_o E) = (ordinal_oa E).
  by move => h; rewrite h in sa;apply/ordinal_pr1.
have g1: sgraph (ordinal_o E) by move: (ordinal_o_or E) => [].
have g2: sgraph (ordinal_oa E) by move => t /Zo_S /setX_P [].
apply:sgraph_exten => // u v. rewrite /ordinal_oa.
split => /graph_on_P1 [uE vE h]; apply /graph_on_P1; split => //.
  case: (equal_or_not u v) => euv; [by right | left].
  by rewrite -(sb _ vE); apply/segmentP; split => //; apply/ordo_leP.
case: h; [ move => ub; exact (proj1(Ha _ _ vE ub)) |  by move ->].
Qed.

(** A [numeration] is a functional graph [f] such that [f(x)] 
is the set of all [f(y)] 
where [y<x] for [r]. A set is [numerable] if well-ordered and there is a 
[numeration]. 
*)


Definition numeration r f:=
  [/\ fgraph f, domain f = substrate r &
  forall x, inc x (domain f) ->  Vg f x = direct_image f (segment r x)].
Definition  numerable r :=
   worder r /\ exists f, numeration r f.

(** Each set has at most one numeration *)
Lemma direct_image_exten f f' a:
  fgraph f -> fgraph f' -> (forall x, inc x a -> Vg f x = Vg f' x) ->
  sub a (domain f) -> sub a (domain f') ->
  direct_image f a = direct_image f' a.
Proof.
move=> fgf fgf' sVVg adf adf'.
set_extens x; move /dirim_P => [u us Jg]; apply /dirim_P; ex_tac.
  move: (pr2_V fgf Jg); aw; move=> ->; rewrite (sVVg _  us). 
  by apply: fdomain_pr1 =>//;  apply: adf'. 
move: (pr2_V fgf' Jg); aw; move=> ->; rewrite -(sVVg _  us). 
by apply: fdomain_pr1 =>//;  apply: adf. 
Qed.

Lemma numeration_unique r f f':
  worder r -> numeration r f -> numeration r f' -> f = f'.
Proof.
move=>  [or wor] [fgf df Vf][fgf' df' Vf'].
set A :=Zo (substrate r)(fun x=> Vg f x <> Vg f' x).
case: (emptyset_dichot A)=> neA.
  apply: (fgraph_exten fgf fgf'); rewrite df => // a xdf.
  by ex_middle nVV; empty_tac1 a; apply: Zo_i.
have sAs: sub A (substrate r) by apply: Zo_S.
move: (wor _ sAs neA)=> [y []]; aw => yA leyA.
move: (sAs _ yA)=> ysr.
have srp: forall z, inc z (segment r y) -> Vg f z = Vg f' z.
  move=> z /segmentP zy; ex_middle nVV. 
  have zA: (inc z A) by  apply: Zo_i=>//; order_tac.
  move: (iorder_gle1 (leyA _ zA)) => yz; order_tac.
have nyA: Vg f y = Vg f' y.  
  rewrite df in Vf; rewrite df' in Vf'. 
  rewrite Vf // Vf' //.
  apply: direct_image_exten =>//.
    rewrite df;  apply: sub_segment.
    rewrite df';  apply: sub_segment.
by move: yA => /Zo_P [_ []]. 
Qed.

(** A numeration of a segment is obtained by restriction *)

Lemma sub_numeration r f x:
  let r' := induced_order r (segment r x) in
  worder r ->  numeration r f -> inc x (substrate r) ->
  numeration r' (restr f (segment r x)).
Proof.
move=> r' wor num xsr.
have sss:= (@sub_segment r x).
have wor':=(induced_wor wor sss).
have [fgf df pf]:= num.
have sr':substrate r' = (segment r x) by rewrite /r'; aw; fprops.
have ssd: sub (segment r x) (domain f) by ue.
have dr:domain (restr f (segment r x)) = substrate r' by bw.
split;bw;fprops => a asr.
have sra: (segment r' a = segment r a).
  apply: segment_induced_a =>//; fprops. 
  apply: segment_segment; fprops.
rewrite sra (restr_ev _ asr) pf; last by apply:ssd.
have ssad: sub (segment r a) (domain f) by rewrite df; apply:sub_segment.
have sax: sub (segment r a) (segment r x).
  move=> t; move: asr => /segmentP p1  /segmentP p2. 
  have or: order r by fprops.
  apply /segmentP;order_tac.
apply: direct_image_exten; bw; fprops.
by move=> z zs; rewrite restr_ev => //; apply: sax. 
Qed.

(** If all segments are numerable so is the whole set *)

Lemma segments_numerables r:
  let r' := fun x => induced_order r (segment r x) in
  worder r -> (forall x, inc x (substrate r)-> numerable (r' x)) ->
  numerable r.
Proof.
move=> r' wor aln.
set f:= (fun x =>  choose (fun y=> numeration (r' x) y)).
have fp: forall x, inc x (substrate r) -> numeration (r' x) (f x). 
  by move=> x xsr; move: (aln _ xsr)=> [_ c];rewrite /f;apply: choose_pr.  
have or: order r by fprops.
have restrp: forall x y, glt r y x ->
   f y = restr (f x) (substrate (r' y)).
  move=> x y ryx.
  have sss: sub (segment r x) (substrate r) by apply: sub_segment.
  have ssy: sub (segment r y) (substrate r) by apply: sub_segment.
  have p6: worder (r' y) by apply: induced_wor =>//.
  apply: (numeration_unique p6); first by apply:  fp; order_tac.
  have <-: (segment (r' x) y) = substrate (r' y).
    rewrite /r'; aw; try apply: sub_segment; rewrite segment_induced //.
  have p1: worder (r' x) by rewrite /r'; apply: induced_wor =>//.
  have p2: numeration (r' x) (f x) by apply: fp; order_tac.
  have p3: inc y (substrate (r' x)) by rewrite /r'; aw; apply /segmentP.
  move: (sub_numeration p1 p2 p3).
  have -> //: induced_order (r' x) (segment (r' x) y) = r' y.
  rewrite /r' segment_induced // iorder_trans //.
  apply: (segment_monotone or (proj1 ryx)). 
have gp: forall x y, glt r y x -> range (f y) = Vg (f x) y. 
  move=> x y yx.
  have xsr: (inc x (substrate r)) by order_tac.
  have ysr: (inc y (substrate r)) by order_tac.
  move: (fp _ xsr) => [fg df vf].  
  move: df; rewrite /r' ; aw => df; last by apply: sub_segment.
  have ysrx: (inc y (segment r x)) by  apply /segmentP.
  rewrite df in vf; rewrite (vf _ ysrx)  (restrp _ _ yx).
  have ->: (segment (r' x) y) = (substrate (r' y)).
    rewrite /r'; aw; [ by rewrite segment_induced | apply: sub_segment].
  have gr: sgraph (restr (f x) (substrate (r' y))) by fprops.
  have fgr: fgraph (restr (f x) (substrate (r' y))) by fprops.
  have hh: sub (substrate (r' y)) (domain (f x)). 
    rewrite df; rewrite /r' (iorder_sr or (@sub_segment r y)). 
    apply:(segment_monotone or (proj1 yx)).
  set_extens u.
    move /(range_gP fgr); bw; move => [z pa pb]; apply /dirim_P; exists z => //.
    by move: pb; bw => ->; apply:(fdomain_pr1 fg); apply: hh.
  move /dirim_P => [z za zb];apply /(range_gP fgr); exists z; bw.
  move: (pr2_V fg zb); aw.
split =>//; exists (Lg (substrate r) (range \o f)).
red; bw; split;fprops.
move=> x xsr; bw.
move: (fp _ xsr) => [ff df vf].
 move: df; rewrite /r' ; aw => df; last by apply: sub_segment.
have gf: sgraph (f x) by fprops.
have p0: sub (segment r x) (substrate r) by apply: sub_segment.
have fgL: fgraph (Lg (substrate r) (range \o f)) by fprops.
set_extens t.
  move=> /(range_gP ff) [z pa pb]; apply /dirim_P; exists z; first ue.
  move: pa; rewrite df => /segmentP h; rewrite pb - (gp _ _ h); apply: Lg_i.
  order_tac.
move /dirim_P => [z xs JL]; move: (p0 _ xs) => zs.
move /segmentP: (xs) => lt1. 
move: (pr2_V fgL JL); aw; bw  => -> /=; rewrite (gp _ _ lt1). 
by apply /(range_gP ff); exists z => //; ue.
Qed. 

(** Any well-ordered is numerable *)

Lemma worder_numerable r:
  worder r ->  numerable r.
Proof.
move=> wor; apply: segments_numerables =>// x xsr.
set r' := fun x => induced_order r (segment r x).
set (A:= Zo (substrate r)(fun z=> ~numerable (r' z))).
case: (p_or_not_p (numerable (induced_order r (segment r x)))) =>// h.
have sA: (sub A (substrate r)) by apply: Zo_S.
have neA: nonempty A  by exists x; apply/Zo_P.
move: (wor)=> [or aux]; move: (aux _ sA neA)=> [y []]; aw=> ys yp.
have rec: forall z, glt r z y -> numerable (r' z).
  move=>z zy; case: (p_or_not_p (numerable (r' z))) =>// nnz.
  have zA: (inc z A) by rewrite /A; apply: Zo_i=>//; order_tac.
  move: (iorder_gle1 (yp _ zA)) => yz; order_tac.
move: ys => /Zo_P [ys ny]; case: ny.
have sss: sub (segment r y) (substrate r) by apply: sub_segment.
apply: segments_numerables; first by  apply: induced_wor.
rewrite /r'; aw; move=> z zy.
have rzy: glt r z y by apply /segmentP.
rewrite segment_induced // iorder_trans //;first by apply: rec.
apply: segment_monotone =>//; order_tac.
Qed.

(** The previous result is obvious by [transfinite_defined] *)

Lemma worder_numerable_bis  r:
  worder r -> numeration r (graph (ordinal_iso r)).
Proof. 
move=> wor.
move: (transfinite_defined_pr target wor) => [suf sf vf].
have ff:= proj1 suf. 
hnf; rewrite (f_domain_graph ff) sf; split; fprops.
move=> x xsr; move: (vf _ xsr); rewrite /Vf; move=> ->; apply:corresp_t.
Qed.


End VonNeumann. 


(** We start with some results that are not in the main files *)

Module Misc2.


(* alternate proof of : there is no set of all cardinals *)
Lemma cantor_bis_alt: non_coll cardinalp.
Proof.
move=> [a ap].
set (s:= union a).  
set (e:= cardinal s).
have cp:= CS_pow \2c e.
case: (cltNge (cantor (CS_cardinal s))).
have /sub_smaller : (sub (\2c ^c e) s) by apply: setU_s1; apply/ ap.
by rewrite (card_card cp).
Qed.


(** Exercise 5.3 of Bourbaki (see part 1)  *)

Lemma exercise5_3 X i j (I:= domain X) (n := cardinal I)
   (ssI := fun k => subsets_with_p_elements k I)
   (uH := fun H => unionb (restr X H))
   (iH := fun H => intersectionb (restr X H))
   (iuH := fun k => intersectionf (ssI k) uH)
   (uiH := fun k => unionf (ssI k) iH):
   fgraph X -> natp n -> natp i -> natp j ->
   (((i = j \/ j <> \0c) -> i +c j <=c csucc n -> sub (iuH i) (uiH j))
   /\ (csucc n <=c i +c j -> (i = j \/ j <=c n) -> sub (uiH i) (iuH j))).
Proof.
move => fgX nN iN jN.
split.
  move => spec le1 t hyp.
  case: (equal_or_not j \0c) => jnz. 
    have i0: i = \0c by case: spec => // ->. 
    move: hyp;  rewrite /iuH /uiH /ssI jnz i0.
    have re: (restr X emptyset) = emptyset.
      apply /set0_P => z /funI_P [a []]; case; case.
    have ->: (subsets_with_p_elements \0c I) = singleton emptyset.
      apply set1_pr.
        by apply/Zo_P; rewrite cardinal_set0;split => //; apply/setP_P; fprops.
      move => z /Zo_P [_]; apply:card_nonempty.
    by rewrite setUf_1 setIf_1 /uH /iH re setUb_0 setIb_0.
  set A := Zo I (fun z => ~ inc t (Vg X z)); set B := I -s A.
  have AI: sub A I by apply : Zo_S.
  move: (cardinal_setC2 AI); rewrite -/n -/B.
  rewrite -(csum2cr A) -(csum2cl A) => eq1.
  have bp: forall z, inc z B -> inc t (Vg X z).
    move => z /setC_P [zI] /Zo_P h; ex_middle tv; case: h;split => //.
  move: (nN);rewrite eq1  => nN'.
  have caN:= (Nat_in_suml (CS_cardinal A) nN'). 
  have cbN :=(Nat_in_sumr (CS_cardinal B) nN').
  case: (NleT_el jN cbN) => jcb.
    have [C [CB cC]]: exists C, sub C B /\ cardinal C = j.
       move: jcb; rewrite -(card_nat jN) => jcb.
       move: (proj2 (eq_subset_cardP _ _)  (cardinal_le_aux1 jcb)).
       by move => [C c1 /card_eqP/esym c2]; exists C. 
     have CI: sub C I by move => s sC; move:(CB _ sC);move /setC_P => [].
     apply /setUf_P; exists C.
       by apply/ Zo_P; split => //; apply/setP_P.
     apply /setIb_P. 
       apply/domain_set0P;bw; apply /nonemptyP => ce.
       by move: cC; rewrite ce cardinal_set0; apply:nesym.
     by bw;move => l lC; bw; apply: bp; apply: CB.
  have: n <c cardinal A +c j.
    rewrite eq1; rewrite csumC (csumC _ j); apply:csum_Mlteq => //.
  move /(cleSltP nN) => le2.
  move: (cleT le1 le2); rewrite csumC (csumC _ j) => eq2.
  move:(csum_le2l jN iN caN eq2) => icA.
  have [C CB cC]: exists2 C, sub C A & cardinal C = i.
     move: icA; rewrite -(card_nat iN) => icA.
     move: (proj2 (eq_subset_cardP _ _)  (cardinal_le_aux1 icA)).
     by move => [C c1 /card_eqP /esym c2]; exists C. 
  have CI: sub C I by move => s sC; move:(CB _ sC); apply: AI.
  have dsi:inc C (ssI i) by apply /Zo_i => //; apply /setP_P.
  move:(setIf_hi hyp dsi)=> /setUb_P [y]; rewrite restr_d.
  by move => yC; bw; move: (CB _ yC) => /Zo_P [].
move => le hij.
move => t /setUf_P [y pa pb]; apply /setIf_P. 
  case: hij => jn; first by rewrite - jn; exists y.
  have kN:= (NS_le_nat jn nN).
  rewrite /ssI; apply /nonemptyP => he.
  move: (subsets_with_p_elements_pr nN kN (refl_equal n)); rewrite he. 
  by rewrite cardinal_set0; move: (binom_pr3 nN kN jn).
move => z /Zo_P[/setP_P szi cz].
move /Zo_P:pa => [/setP_P syi cy].
case: (emptyset_dichot (y \cap z)) => di.
  move: (sub_smaller (setU2_12S syi szi)); rewrite (csum2_pr5 di). 
  rewrite -(csum2cr y) cz -(csum2cl y) cy -/n.
  move => le1; case: (cleNgt (cleT le le1) (cltS nN)).
move: di => [l /setI2_P [jz jy]].
apply/setUf_P; exists l; bw.
have jr: inc l (domain (restr X y)) by bw.
move: (setIf_hi pb jr); bw.
Qed.



(** Alternate proof of Cantor Bernstein. 
First, any increasing function on [powerset E] has a fix-point.
Second, assume [f: E-> F] and [g: F-> E] are injections. 
There is a set [M] such that [E-M = g(F- f(M)) ]. 
The function that maps [t] to [f(t)] if [t] is not in this set, 
to the unique [y] in [F-f(M)] such that [t = g(y)] is a bijection [E-> F].
 *)

Lemma Cantor_Bernstein_aux E (g: fterm) 
  (m:= union (Zo (powerset E)(fun x=> sub x (g x)))):
  (forall x, sub x E -> sub (g x) E) ->
  (forall x y, sub x y -> sub y E -> sub (g x) (g y)) ->
  (sub m E /\ g m = m).
Proof. 
move=> p1 p2; rewrite /m; set (A := Zo _ _).
have su: (sub (union A) E). 
  by apply: setU_s2 => y /Zo_P [q1 q2]; apply /setP_P.
have p3: (forall x, inc x A -> sub x (g (union A))). 
  move=> x /Zo_P [q1 q2];apply: (sub_trans q2); apply: p2 =>//.
  by apply setU_s1;  apply: Zo_i. 
have p4:  (sub (union A) (g (union A))). 
  by move=> t /setU_P [y ty yA]; move: (p3 _ yA) =>q; apply: q.
split => //; apply: extensionality => //. 
move=> t tg;move: (p1 _ su) => p5; move: (p2 _ _ p4 p5) => p6.
by apply: (setU_i tg); apply: Zo_i => //; apply /setP_P.
Qed.


Lemma Cantor_Bernstein2_full f g
  (E:= source f) (F:= source g)
  (h:= fun x => E -s (Vfs g (F -s (Vfs f x))))
  (m:= union (Zo (powerset E) (fun x : Set => sub x (h x))))
  (T:= Vfs g (F -s (Vfs f m))) 
  (p := fun a y => [/\ inc y F, ~ inc y (Vfs f m) & a = Vf g y])
  (f2:= Lf (fun a =>Yo (inc a T) (select (p a) F) (Vf f a)) E F):
  injection f -> injection g -> source f = target g -> source g = target f ->
  bijection_prop f2 (source f)(source g).
Proof. 
move=> [ff injf] [fg injg] sf sg.
have p1:  (forall x, sub x E -> sub (h x) E) by move=> x xE; apply: sub_setC.
have p2: (forall x y, sub x y -> sub y E -> sub (h x) (h y)). 
  move=> x y xy yE; rewrite /h => t /setC_P [tE ti]. 
  apply /setC_P;split => //; dneg aux.
  have q1: sub x (source f) by apply: sub_trans yE. 
  have pa: sub (F -s Vfs f y) (source g) by apply:sub_setC.
  have pb: sub (F -s Vfs f x) (source g) by apply:sub_setC.
  move /(Vf_image_P fg pa): aux => [u] /setC_P [uF nuy] Wu.
  apply /(Vf_image_P fg pb); exists u => //;apply /setC_P;split =>//; dneg aux2.
  move /(Vf_image_P ff q1): aux2 => [s sx ->]; apply /(Vf_image_P ff yE).
  by exists s=> //; apply: xy.
move: (Cantor_Bernstein_aux p1 p2) => [].
rewrite -/m /h -/T => mE hm.
have sc: sub (F -s (Vfs f m)) (source g) by apply: sub_setC.
set Pa := (Vf_image_P fg sc); set Pb := (Vf_image_P ff mE).
have TE: (sub T E) by move=> t /Pa [u] /setC_P [uF _] ->; rewrite /E sf; Wtac.
move: (setC_K TE); rewrite hm => cEm.
have sp: forall a, inc a T -> p a (select (p a) F).
  move => a aT; apply: (proj1 (select_pr _ _)).
    by move: aT => /Pa [u] /setC_P [uF num] Wu;ex_tac. 
  move => // x y xF [_ _ xa] yF [_ _ ya].
  by apply: injg => //; rewrite -xa -ya.
rewrite /f2/bijection_prop; aw; split => //; apply: lf_bijective.
+ move=> z zE /=;Ytac zT; [ exact (proj31 (sp _ zT)) | rewrite /F; Wtac].
+ move=> u v uE vE; rewrite /f2; Ytac uT; Ytac vT=> aux.
  - by move: (sp _ uT) (sp _ vT)=> [_ _ W1] [_ _ W2]; rewrite W1 W2 aux.
  - move: (sp _ uT) => [f1u  nf1i Wf1] ; case: nf1i; apply /Pb.
    by rewrite aux;exists v => //; rewrite -hm;apply /setC_P.
  - move: (sp _ vT) => [f1u nf1i Wf1]; case: nf1i; apply /Pb.
    by rewrite -aux;exists u => //; rewrite -hm;apply /setC_P.
  - by apply: injf.
+ move=> y yF; case: (inc_or_not y (Vfs f m)).
    move /Pb => [u um Wy]; exists u; first by apply: (mE). 
    by rewrite /f2 Y_false //;  move: um; rewrite - hm => /setC_P [].
  move=> aux; exists (Vf g y); first by rewrite /E; Wtac.  
  have wt: inc (Vf g y) T by apply /Pa; exists y => //; apply /setC_P. 
  rewrite /f2; Ytac0; move: (sp _ wt) => [q1 q2 q3].
  by symmetry; apply: injg.
Qed.

Lemma Cantor_Bernstein2 f g:
  injection f -> injection g -> source f = target g -> source g = target f ->
  equipotent (source f)(source g).
Proof. 
move => pa pb pc pd.
move:(Cantor_Bernstein2_full pa pb pc pd).
by set f0 := Lf _ _ _ => fp; exists f0.
Qed.



(** Let [E] be an ordered set and [ssub F] the set of strict upper bounds of [F].
Cantor says that [E] is well-ordered if [E] has a least element and, for any 
non-empty subset [F], if [ssub F] is not empty, it has a least element. 
This agrees with the standard definition *)

Lemma cantor_worder r 
   (ssub := fun E => Zo (substrate r)(fun z => forall x, inc x E -> glt r x z)):
   order r -> nonempty (substrate r) -> 
   ( worder r <->
    ( (has_least r)
     /\ (forall F, sub F (substrate r) -> nonempty F -> nonempty (ssub F) ->
       (has_least (induced_order r (ssub F)))))).
Proof.
move => or nesr; split.
  move => [_ wor ]; split;first by apply: worder_least.
  move => F sfr nef nef1; apply: wor => //; apply: Zo_S.
move => [[y [ysr ly]] pb]; split => //.
move => x xsr [x0 x0x].
case: (inc_or_not y x).
  move => yx; exists y; red; rewrite (iorder_sr or xsr); split => //.
  by move => t tx; apply /iorder_gleP => //; apply: ly; apply: xsr.
move => nyx.
set F:= Zo (substrate r)  (fun z => forall t, inc t x -> glt r z t).
have sf1: sub F (substrate r) by apply: Zo_S.
have yf1: inc y F.
  apply: Zo_i => // t tx; split; [ apply: (ly _ (xsr _ tx)) | dneg yt; ue].
set G := ssub F.
have pa : sub x G.
  move => t tx; apply: Zo_i; first by apply: xsr.
  by move => tv /Zo_P [_ aux]; apply: aux.
have neG: nonempty G by exists x0; apply: pa.
have neF: nonempty F by exists y.
have sg: sub (ssub F) (substrate r) by apply: Zo_S.
move: (pb _ sf1 neF neG) => [a []]; rewrite (iorder_sr or sg) => aG h.
exists a; hnf; rewrite (iorder_sr or xsr).
have qa: forall t, inc t x -> gle r a t.
  move => t tx.
  move: (h _ (pa _ tx)) => cp1; exact (iorder_gle1 cp1).
case: (inc_or_not a x) => iax.
  by split => // t tx; apply /iorder_gleP => //; apply: qa.
have aF: inc a F.
  apply: Zo_i; [by apply: sg | move => t tx; split;[ by apply: qa| dneg ta;ue]].
by move: aG => /Zo_P [_ p]; move: (p _ aF) => [_].
Qed.


(* variant of existence of division and substraction *)
Lemma ord_div_nonzero_b a b c:
  ordinalp a -> ordinalp b -> ordinalp c ->
  a <o (b *o c)  -> b <> \0o.
Proof.
move=> oa ob oc ale bz.
by move: ale; rewrite bz oprod0l => /olt0.
Qed.


Lemma ord_div_nonzero_b_bis  a b:
  ordinalp a -> \0o <o b ->
  exists2 c, ordinalp c & a <o (b *o c).
Proof.
move=> oa bp.
move: (oprod_M1le bp (OS_succ oa)) (oltS oa);rewrite  -(osucc_pr oa) => p1 p2.
exists (a +o \1o); [ exact (proj32_1 p2) | exact (olt_leT p2 p1) ].
Qed.

Lemma odivision_exists_alt a b c:
  ordinalp a -> ordinalp b -> ordinalp c ->
  a <o (b *o c) ->
  exists q r, odiv_pr1 a b c q r.
Proof.
move=> oa ob oc [] /ordinal_le_P0; rewrite /oprod2.
set pbc:= (order_prod2 _ _) => qa qb.
move: (conj qa qb) => a_bc; clear qa qb.
have obc: worder pbc by rewrite /pbc; fprops.
move: (ordinal_lt_pr2 obc a_bc) => [f [x [xsp rgx om]]].
have orb: order (ordinal_o b) by fprops.
have orc: order (ordinal_o c) by fprops.
have ora: order (ordinal_o a) by fprops.
move: (ordinal_o_sr b) (ordinal_o_sr c) => sb sc.
move: (xsp); rewrite orprod2_sr // setX2_P.
rewrite sb sc;move=> [fgx dx xa xb].
have [w wsb wle]: exists2 w, inc w b &
  (forall t, inc t b -> gle (ordinal_o b) w t).
  have nesb: nonempty (substrate (ordinal_o b)) by rewrite sb;ex_tac.
  move:  (worder_least  (ordinal_o_wor ob) nesb)=> [w [p1 p2]]. 
  rewrite sb in p1 p2; ex_tac.  
set y := variantLc  (Vg x C0) w.
have ysp: inc y (substrate pbc).
  rewrite orprod2_sr //; apply /setX2_P; rewrite sb sc /y; bw; split;fprops.
set r := (Vg x C1) ; set q := (Vg x C0).
have lt_rb: r <o b by apply /oltP0; split => //; exact (ordinal_hi ob xb).
have lt_qc: q <o c by apply /oltP0; split => //; exact (ordinal_hi oc xa).
have oq:= proj31_1 lt_qc.
have odr:= proj31_1 lt_rb.
have worr: worder (ordinal_o r) by fprops.
have orq: order (ordinal_o q) by fprops.
have sq :=(ordinal_o_sr q).
have sqc: sub q c by move:  lt_qc => [[_ _ h] _].
have srb: sub r b  by move:  lt_rb => [[_ _ h] _].
exists q, r; split => //; split => //.
have rsub: forall r A u v, order r -> sub A (substrate r) ->
  (gle (induced_order r A) u v <-> [/\ inc u A, inc v A & gle r u v]).
  move=> r0 A0 u v or Asr; split.
    move /setI2_P => [pa] /setXp_P [ua va];split => //.
  by move=> [uA vA rr]; apply /iorder_gleP.
set A:= segment pbc y.
have iuA: A = product2 q b.
  have pa: inc C0 (doubleton C0 C1) by fprops.
  have pb: inc C1 (doubleton C0 C1) by fprops.
  set_extens u.
    move /segmentP => [] /(orprod2_gleP orb orc).
    rewrite sb sc variant_V_ca variant_V_cb.
    move=> [pc v2p h] unv.
    move: pc => /setX2_P [fgu du ua ub]; apply /setX2_P => //.
    case: h;last by move /(ordo_ltP oc ua xa).
    move=> [r1 r2]; move: (wle _ ub) => r3.
    have Wv: w =  (Vg u C1) by order_tac.
    case: unv;rewrite /y;apply: fgraph_exten => //; [fprops | bw |].
    rewrite du; move=> x0 x0d; try_lvariant x0d; ue.
  move /setX2_P => [fgu du ua ub].
  have uc:inc (Vg u C0) c by apply: sqc.
  apply /segmentP; split; last first.
     move => uy; case: (ordinal_irreflexive oq).
     move: ua; rewrite uy /y variant_V_ca //.
  apply  /(orprod2_gleP orb orc); rewrite sb sc;split => //.
  + by apply /setX2_P. 
  + rewrite /y;apply /setX2_P;bw;split;fprops.
  + right; rewrite variant_V_ca; apply /(ordo_ltP oc uc) => //.
have ss3: substrate pbc = product2 c b by rewrite orprod2_sr // sb sc.
have ss2: sub A (substrate pbc) by apply: sub_segment.
have sA: sub A (product2 c b) by ue.
have wo2: worder (induced_order pbc A) by apply: induced_wor.
have obq: ordinalp (b *o q) by fprops.
have ->:b *o q = ordinal (induced_order pbc A).
  symmetry; apply: ordinal_o_isu2 => //. 
  set r1 :=  (order_prod2 (ordinal_o b) (ordinal_o q)).
  have wo1: worder r1 by rewrite /r1; fprops.
  apply: orderIT (ordinal_o_is wo1).
  have or2:order (induced_order pbc A) by fprops.
  have or1:order r1 by fprops.
  suff: r1 = (induced_order pbc A) by move=> ->; apply: orderIR.
  apply: order_exten => // u v; split.
    move /(orprod2_gleP orb orq);rewrite sb sq; move => [pa pb pc].
    apply /iorder_gleP; try ue; apply /(orprod2_gleP orb orc).
    rewrite sb sc; split => //; try( apply :sA ; ue).
    case: pc; [ by left | move => [ /ordo_leP [sa ssb ssc] sd] ; right ].
    split => //;apply /ordo_leP;split;fprops.
  move => /iorder_gle5P [uA vA] /(orprod2_gleP orb orc) [_ _ h].
  apply /(orprod2_gleP orb orq); rewrite sb sq;split => //; try ue.
  case: h; [by left | move=> [/ordo_leP [sa ssb ssc] sd]; right].
  move: uA vA;rewrite iuA;move /setX2_P => [_ _ uq _] /setX2_P [_ _  vq _].
  split => //;apply /ordo_leP;split;fprops.
set C := induced_order pbc (segment pbc x).
have oraC: order_isomorphic C (ordinal_o a).
  have injf:= (order_morphism_fi om).
  move: om=> [o1 o2 [ff sf tf] mof].
    apply: orderIS; rewrite /C -rgx.  
  have sr: sub (range (graph f)) (substrate pbc).
     rewrite - tf; fprops; apply: f_range_graph =>//.  
  have o3:= (proj1(iorder_osr o2 sr)).
  have aux: (source (restriction1 f (source f))) = source f.
     rewrite  /restriction1; aw.
  exists (restriction1 f (source f)); split => //.
    red; rewrite - sf; aw; split => //; first by apply: restriction1_fb.
    rewrite /restriction1; aw; rewrite -image_by_fun_source //.
  red; rewrite aux => u v usf vsf;rewrite restriction1_V // restriction1_V //.
  rewrite /pbc mof // rsub //; split; last by move => [_ _].
  move=> Wuv; split => //; Wtac. 
rewrite - (ordinal_o_o oa) -(ordinal_o_o odr).
have opbc: order pbc by fprops.
symmetry;apply: orsum_invariant5 => //. 
apply: orderIT oraC.
set fo:= order_sum2 _ _.
have sfo: substrate fo = canonical_du2 A r.
  by rewrite /fo orsum2_sr ;aw; fprops; rewrite ordinal_o_sr.
have ysp': inc y (product2 c b) by ue.
have xsp': inc x (product2 c b) by ue.
have lexy: gle pbc y x.
  apply /(orprod2_gleP orb orc); rewrite sb sc /y;bw; split => //; left. 
  by split => //;apply: wle.
have ltAx: forall t, inc t A -> glt pbc t x.
   move =>t /segmentP lttxy; order_tac.
have rprop: forall z, inc z r -> inc (variantLc (Vg x C0) z) (product2 c b).
  move=> z zB; apply /setX2_P; bw; split;fprops.
have p1: forall z, inc z r -> gle pbc  y (variantLc (Vg x C0) z).
  move=> z zB; apply /(orprod2_gleP orb orc); rewrite sb sc; split;fprops.
  by rewrite /y;bw;left; split => //; apply: wle; apply: srb.
set g := fun z => Yo (Q z = C0) (P z) (variantLc (Vg x C0) (P z)).
have ssx: sub (segment pbc x) (substrate pbc) by apply: sub_segment.
have srC: substrate C = (segment pbc x) by rewrite /C; aw. 
have ta: forall z, inc z (substrate fo) -> inc (g z) (substrate C).
  move=> z; rewrite sfo; move /candu2P=> [pz xx].
  rewrite srC; apply /segmentP;case :xx.
    by move => [Pz Qz]; rewrite /g; Ytac0;apply: ltAx => //.
  move => [Pz Qz]; rewrite /g Qz; Ytac0; split; last first.
    move => bad; case: (ordinal_irreflexive odr); rewrite {1} /r -bad; bw.
  apply /(orprod2_gleP orb orc); rewrite sb sc.
  move /(ordo_ltP ob (srb _ Pz) xb): (Pz) => [pa pb].
  split; [ by apply: rprop | exact | left].
  by rewrite variant_V_ca variant_V_cb.
have bg: bijection (Lf g (substrate fo) (substrate C)).
  apply: lf_bijective => //.
    move=> u v; rewrite sfo /g; move /candu2P => [pu p3]  /candu2P [pv p4].
    case: p3; case: p4; move=> [p5 p6][p7 p8];rewrite p6 p8; Ytac0; Ytac0=> aux.
    + by apply: pair_exten =>//;  ue.
    + move/segmentP: p7 => lt1;move:(p1 _ p5);rewrite -aux => lt2; order_tac.
    + move/segmentP:p5 => lt1; move:(p1 _ p7); rewrite aux => lt2; order_tac.
    + by apply: pair_exten => //; [ move: (f_equal (Vg ^~ C1) aux); bw | ue ].
  move=> v; rewrite srC => /segmentP hyp.
  case: (p_or_not_p (glt pbc v y)) => aux.
    exists (J v C0); last by rewrite /g; aw; Ytac0.
    rewrite sfo; apply /candu2P; split;fprops;left. 
    by aw; split => //; apply /segmentP.
  move: hyp aux => [] /(orprod2_gleP orb orc); rewrite sb sc.
  move => [vp xp aux] nvx cvy.
  case: aux; last first.
    move=> aux2; case: cvy.
    split; last by rewrite /y;move: aux2 =>[_ vx]; dneg ww; rewrite ww; bw.
    apply /(orprod2_gleP orb orc);  rewrite sb sc.
    by split => //;rewrite /y; bw; right.
  move=> [sva leva].
  have vpr: v = variantLc (Vg x C0) (Vg v C1).
    move: vp => /setX2_P [a1 a2 a3 a4].
    apply: fgraph_exten => //;[ fprops | bw |]. 
    rewrite a2;  move => z zd; try_lvariant zd.
  have ltva:glt (ordinal_o b) (Vg v C1) (Vg x C1).
    split => //; clear cvy; dneg ww; rewrite vpr ww.
    symmetry; move:xp => /setX2_P [a1 a2 _ _].
    apply: fgraph_exten => //;[ fprops | bw |]. 
    rewrite a2;  move => z zd; try_lvariant zd.
  exists (J  (Vg v C1) C1); last by rewrite /g; aw; rewrite -vpr; Ytac0.  
  rewrite sfo;apply: candu2_prb => //.
  move /ordo_leP: leva => [pa pb pc].
  by apply /(ordo_ltP ob pa pb).
exists (Lf g (substrate fo) (substrate C)).
  apply: total_order_isomorphism; rewrite ? lf_source ? lf_target //.
    rewrite /fo;apply: orsum2_totalorder => //.
      apply: total_order_sub => //;apply: worder_total => //.
      apply: worder_total => //.
  apply: (proj1 (iorder_osr opbc ssx)).
rewrite sfo; rewrite sfo in ta.
have oiA: order (induced_order pbc A) by fprops.
have oor:  order (ordinal_o r) by fprops.
move=> u v uc vc /=; rewrite lf_V // lf_V // /fo /C rsub //.  
move /orsum2_gleP => [_ _ h]; rewrite - srC;split => //; try (apply: ta => //).
rewrite /g; case: h. 
    move=> [h1 h2 h3]; Ytac0; Ytac0; apply: (iorder_gle1 h3).
  move=> [h1 h2 h3];  Ytac0; Ytac0.
  apply /(orprod2_gleP orb orc); rewrite sb sc; bw.
  move /ordo_leP: h3 => [h3 h4 h5].
  split; [by apply: rprop | by apply: rprop | left; split => //].
  by apply /ordo_leP; split => //;  apply: srb.
move => [h1 h2]; Ytac0; Ytac0.
move/candu2P: uc => [_]; case; last by move => [_ bad]; case: C1_ne_C0; ue.
move => [/segmentP [pa _] _].
move/candu2P: vc => [_];case; [ by move=> [_] | move=> [/p1 h0 _]; order_tac].
Qed. 

Lemma odiff_pr_alt a b 
  (c := ordinal (induced_order (ordinal_o b) (b -s a))):
  a <=o b -> (ordinalp c /\ b = a +o c).
Proof.
move => [oa ob ab].
set F:= b -s a.
set B := ordinal_o b.
have wB: worder B by rewrite /B;fprops.
move: (ordinal_o_sr b); rewrite -/B => sB.
have Es: sub a (substrate B) by ue.
have Fs: sub F (substrate B) by rewrite sB; apply: sub_setC.
set r := induced_order B a.
set r' := induced_order B F.
have wor: worder r by apply: induced_wor.
have wor': worder r' by apply: induced_wor.
have -> : c = ordinal r'  by done.
set A:= ordinal r; set C := ordinal r'.
have oA: ordinalp A by apply: OS_ordinal.
have oC: ordinalp C by apply: OS_ordinal.
split; first by exact.
have ra : r = ordinal_o a.
  apply: order_exten; fprops.
  move => x y; split => le1.
    move: (iorder_gle3 le1) => [p1 p2].
    move: (iorder_gle1 le1) => /ordo_leP [_ _ h]; apply /ordo_leP;split => //.
  move /ordo_leP: le1 => [pa pb pc]; apply /iorder_gleP => //.
  apply /ordo_leP;split;fprops.
have <-: A = a by rewrite /A ra; apply: ordinal_o_o.
rewrite /A/C - (ordinal_o_o ob).
symmetry; apply: orsum_invariant5 => //.
have orb: order B by fprops.
have or: order r by fprops.
have or': order r' by fprops.
have sr: substrate r = a by rewrite /r; aw.
have sr': substrate r' = F by rewrite /r'; aw.
have ta: forall x, inc x (substrate (order_sum2 r r')) 
   -> inc (P x) (substrate B).
  move=> x; rewrite orsum2_sr // => /candu2P; rewrite sr sr'.
  by move=> [_ ]; case; move=> [pqx _]; [apply: Es |apply: Fs].
exists (Lf P  (substrate (order_sum2 r r')) (substrate B)).
have xxa: total_order (order_sum2 r r').
  by apply: orsum2_totalorder; apply: worder_total.
have xxb:bijection (Lf P (substrate (order_sum2 r r')) (substrate B)).
  apply: lf_bijective => //.
    move=> x y;rewrite orsum2_sr //;move /candu2P=> [px h1] /candu2P [py h2] sp.
    apply: pair_exten =>//.
    case: h1; case: h2; move=> [Px Qx][Py Qy];rewrite Qx Qy //.
      by move: Px Py; rewrite sr sr' /F - sp;  move /setC_P => [].
      by move: Px Py; rewrite sr sr' /F - sp => aux; move /setC_P => [].
  move=> y ysb; case: (inc_or_not y a) => yE.
    exists (J y C0); last by aw. 
    by rewrite orsum2_sr //; apply candu2_pra; rewrite sr.
  have yF: inc y F by rewrite /F - sB; apply /setC_P; split.
  exists (J y C1); last by aw. 
  by rewrite orsum2_sr //; apply :candu2_prb; rewrite sr'.
apply: total_order_isomorphism => //; aw.
move=> x y xsr ysr /=; rewrite lf_V //  lf_V //.
move: xsr ysr;  rewrite orsum2_sr //; move=> xsr ysr => ha.
move /orsum2_gleP:ha => [_ _ ]; case.
    by move=> [_ _ h1]; move: (iorder_gle1 h1).
  by move=> [_ _ h1]; move: (iorder_gle1 h1).
move => [qxa qyb].
move: xsr ysr => /candu2P [px h1] /candu2P [py h2].
case: h1; last by move=> [_];  rewrite qxa => aux; case: C0_ne_C1.
case: h2; first by move=> [_ xx]; rewrite xx in qyb; case: qyb. 
rewrite sr sr'; move=> [Py _] [Px _ ].
move: (worder_total wB)=> [_ tob].
have xb: inc (P x) (substrate B) by apply:  Es.
have yb: inc (P y) (substrate B) by apply:  Fs.
case: (tob _ _ xb yb) => //.
move: Py => /setC_P [_ nya] /ordo_leP [iyb ixb xy]; case: nya.
case:  (ordinal_sub (ordinal_hi ob iyb)  (ordinal_hi ob ixb) xy).
  by move=> ->.
apply: (ordinal_transitive oa Px).
Qed.

Lemma odiff_wrong_alt a b
  (c := ordinal (induced_order (ordinal_o b) (b -s a))):
  b <=o a -> c = \0c.
Proof.
rewrite /c; move=> [oa ob ab];rewrite (setC_T ab).
have oon: order (ordinal_o b) by apply: ordinal_o_or.
apply: ordinal0_pr1; aw; last by fprops.
apply: (proj1 (iorder_osr oon _)); fprops.
Qed.

Lemma osum_limit_alt x y: ordinalp x -> limit_ordinal y -> 
  limit_ordinal (x +o y).
Proof.
move=> ox ly.
move:(proj31 ly) => oy.
case: (ozero_dichot ox) => xz; first by rewrite xz osum0l //. 
move /limit_ordinal_P:ly => [ ynz yl]. 
move: (@osum_Meqlt \0o y x ynz ox); rewrite (osum0r ox) => lt0.
apply /limit_ordinal_P; split; first exact: (olt_ltT xz lt0).
move => z lt1.
have y1: \1o <o y by rewrite - osucc_zero; apply: yl. 
have oz:= proj31_1 lt1.
rewrite - (osucc_pr oz).
case: (oleT_ee oz ox) => zx.
  exact (ole_ltT (osum_Mleeq zx OS1) (osum_Meqlt y1 ox)).
move: (odiff_pr zx) => [pd pe].
rewrite pe in lt1.
move: (yl _ (osum_Meqltr pd oy ox lt1)). 
rewrite - (osucc_pr pd) => pg.
by move: (osum_Meqlt pg ox);  rewrite (osumA ox pd OS1) -pe.
Qed.


(** Injectivity of cardinal successor, according to Bourbaki *)

Lemma succ_injective2 a b: cardinalp a ->  cardinalp b -> 
  a +c \1c = b +c \1c -> a = b.
Proof. 
move=> ca cb hyp.
suff: equipotent a b.
  by move/card_eqP; rewrite (card_card ca) (card_card cb).
move: (disjointU2_pr3 a card_one C1_ne_C0) => eq1. 
move: (disjointU2_pr3 b card_one C1_ne_C0) => eq2.
rewrite -hyp in eq2; clear hyp.
move: (EqS eq1) => [f [bf sf tf]] {eq1}.
move: (EqS eq2)=> [g [bg sg tg]] {eq2}.
set (A0:= Vfs f (a *s1 C0)).
set (B0:= Vfs g (b *s1 C0)).
rewrite /card_one in sf sg.
set xA:= J emptyset C0; set xB:=  J emptyset C1.
have main: (forall c h, bijection h -> source h =
    (c *s1 C0) \cup  ((singleton emptyset) *s1 C1) ->
    let w:= Vfs h (c *s1 C0) in
      let v:= Vf h xB in
       [/\  ~(inc v w),  target h = w \cup (singleton v) & c \Eq w]).
  move=> c h bh sh w v.
  move: (bij_function bh)=> fh. 
  have  sh1:sub (c *s1 C0) (source h) by rewrite sh; apply: subsetU2l.
  have sh2: inc xB (source h).
    rewrite sh; apply: subsetU2r; rewrite /xB;apply /indexed_P;aw;split;fprops.
  split. 
    move /(Vf_image_P fh sh1) => [u vs vw]. 
    have uW: (u = xB).
      by move: bh=> [[ _ ih] _]; apply: ih=>//; apply: sh1.
   move: vs; rewrite uW; move /indexed_P=> [_ _]; rewrite /xB; aw;fprops.
    set_extens t. 
      move:bh => [_ sjh] tt; move:((proj2 sjh) _ tt)=> [y ys hy].
      rewrite sh in ys; case /setU2_P: ys => p.
        apply /setU2_P; left; rewrite -hy; apply  /(Vf_image_P fh sh1); ex_tac.
      have yp: y = xB.
        move:p => /indexed_P; aw; move => [py Py Qy]. 
        by apply: pair_exten; rewrite /xB; aw;fprops; move /set1_P: Py.
      rewrite -hy yp -/v; apply /setU2_P; right; fprops.
    case /setU2_P.
      move /(Vf_image_P fh sh1)=> [u us ->]; Wtac.
    move /set1_P=> ->; rewrite /v; Wtac.
  apply: (EqT (Eq_indexed c C0)).
  exists (restriction1 h (c *s1 C0));split. 
    by apply: restriction1_fb; move: bh => [ih _].
    rewrite /restriction1; aw. 
    rewrite /restriction1; aw. 
apply/card_eqP.
move: (main _ _ bf sf) (main _ _ bg sg); clear main.
simpl; fold A0 B0.
move=> [nWxfA0 tfA /card_eqP ->][nWxfB0 tgB /card_eqP ->].
have tfg: target f = target g by ue.
case: (equal_or_not (Vf f xB) (Vf g xB))=> efg.
  suff: (A0 = B0) by move=> ->.
  set_extens t => ts.  
    have : (inc t (target f)) by rewrite tfA;  apply: subsetU2l.
    rewrite tfg tgB;case /setU2_P =>//. 
    move /set1_P;rewrite -efg =>tv; case: nWxfA0; ue.
  have : (inc t (target f)) by rewrite tfg tgB; apply: subsetU2l.
  rewrite tfA; case /setU2_P =>//. 
  move /set1_P; rewrite efg =>tv; case: nWxfB0; ue.
set (C0:= A0 \cap B0). 
have ->: (A0 = C0 \cup (singleton (Vf g xB))). 
  set_extens t => ts. 
     have : (inc t (target f)) by rewrite tfA;apply: subsetU2l.
     rewrite tfg tgB; case /setU2_P.
     by move=> tB; apply :setU2_1; apply: setI2_i.
     move /set1_P => ->; apply /setU2_P; right; fprops.
  case /setU2_P: ts; first by move/ setI2_P => [].
  move /set1_P => tu.
  have: (inc t (target f)). rewrite tfg  tgB tu; fprops.
  by rewrite tfA; case/setU2_P =>// /set1_P tw; case: efg; rewrite -tu -tw.
have ->: (B0 = C0 \cup (singleton (Vf f xB))).
  set_extens t => ts. 
     have : (inc t (target g)) by rewrite tgB;apply: setU2_1.
     rewrite -tfg tfA; case /setU2_P.
       by move=> tB; apply: setU2_1; apply: setI2_i.
     move /set1_P => ->; apply /setU2_2; fprops.
  case /setU2_P: ts; first by move /setI2_P => [].
  move /set1_P =>  tu.
  have: (inc t (target f)).
       rewrite tfA; apply: setU2_2; aw; ue. 
  rewrite tfg tgB; case /setU2_P =>// /set1_P tw.
  by case: efg; rewrite -tu -tw.
rewrite !csucc_pr //;move /setI2_P => [] //.
Qed.


End Misc2.


(* -------------------------------------------------------------------- *)

Module  ZermeloEq.
(* Equipotency, according to Zermelo *)


Definition zprod a := Zo (powerset (union a)) 
  (fun y=> forall x,  inc x a -> singletonp (y \cap x)). 

Definition zprod2 a b:= zprod (doubleton a b).

Lemma zprod2_P a b y:
  inc y (zprod2 a b) <->
  [/\ sub y (a \cup b), singletonp (y \cap a) & singletonp (y \cap b)].
Proof.
split.
  move => /Zo_P [] /setP_P pa pb;split => //; apply: pb; fprops.
move => [pa pb pc]; apply /Zo_P; split; first by apply /setP_P.
by move => x /set2_P; case => ->.
Qed.

Definition zpr x a := union  (x \cap a).
 
Lemma zprod2_pr1 a b x:
  inc x (zprod2 a b) -> 
  ((x \cap a = singleton (zpr x a)) /\
   (x \cap b = singleton (zpr x b))).
Proof. 
rewrite /zpr.
by move /zprod2_P => [pa [t ->] [s ->]]; rewrite ! setU_1.
Qed.

Lemma zprod2_pr0 a b x:
  inc x (zprod2 a b) ->
  [/\ inc (zpr x a) a, inc (zpr x b) b, inc (zpr x a) x & inc (zpr x b) x].
Proof.
move=>  xz; move: (zprod2_pr1 xz) => [pa pb].
have /setI2_P [sa sb] : (inc (zpr x a) (x \cap a)) by rewrite pa; fprops. 
have /setI2_P [sc sd] //: (inc (zpr x b) (x \cap b)) by rewrite pb; fprops. 
Qed.

Lemma zprod2_pr0aa a b x:
  inc x (zprod2 a b) -> inc (zpr x a) a.
Proof. by move=> xz; case: (zprod2_pr0 xz).  Qed.

Lemma zprod2_pr0ax a b x:
  inc x (zprod2 a b) -> inc (zpr x a) x.
Proof. by move=> xz; move: (zprod2_pr0 xz) => []. Qed.

Lemma zprod2_pr0bb a b x:
  inc x (zprod2 a b) -> inc (zpr x b) b.
Proof. by move=> xz; move: (zprod2_pr0 xz) => []. Qed.

Lemma zprod2_pr0bx a b x:
  inc x (zprod2 a b) -> inc (zpr x b) x.
Proof. by move=> xz; move: (zprod2_pr0 xz) => []. Qed.

Lemma zprod2_pr1a a b x z:
  inc x (zprod2 a b) -> 
  inc z a -> inc z x -> z = zpr x a.
Proof. 
move=> xz za zb; apply/set1_P;move: (zprod2_pr1 xz)=> [<- _]; fprops.
Qed.

Lemma zprod2_pr1b a b x z:
  inc x (zprod2 a b) -> 
  inc z b -> inc z x -> z = zpr x b.
Proof. 
move=>  xz zb zx;  apply/set1_P; move: (zprod2_pr1 xz)=> [_ <-]; fprops.
Qed.

Lemma zprod2_pr2 a b x:
  inc x (zprod2 a b) -> x = doubleton (zpr x a) (zpr x b).
Proof. 
move=> yz; set_extens z; aw.
  move: (yz) => /zprod2_P [pa pb pc]. 
  move=> xy; move: (pa _ xy);case /setU2_P => h.
   rewrite (zprod2_pr1a yz h xy); fprops.
   rewrite (zprod2_pr1b yz h xy); fprops.
by move: (zprod2_pr0ax yz) (zprod2_pr0bx yz) => pa pb;case /set2_P=> ->.
Qed.

Lemma intersection_singletonP a b c:
  (a \cap (singleton b) = singleton c) <->
  (inc c a /\ c = b).
Proof. 
split. 
  move=>  iis; move:(set1_1 c); rewrite -iis => /setI2_P.
  by move => [pa] /set1_P. 
move=> [pa <-]; apply: set1_pr; first by fprops.
by move => t /setI2_P [_] /set1_P.
Qed.

Lemma is_singleton_int a b c:
  inc c a -> inc c b -> (forall u, inc u a-> inc u b -> u = c) ->
  singletonp (a \cap b).
Proof. 
move=> ca cb h;exists c; apply: set1_pr; first by fprops.
by move => t /setI2_P [pa pb]; apply : h.
Qed.

Lemma zprod_singleton M r: ~ inc r M ->
  let N := zprod2 M (singleton r) in
    ( (forall u, inc u M -> inc (doubleton u r) N) /\
      (forall x, inc x N -> exists2 u, inc u M & x = doubleton u r)).
Proof.
move=> nrM N; split.
  move=> u uM; apply /zprod2_P;split => //. 
      move=> t /set2_P; case => ->; apply /setU2_P; [left | right]; fprops.
    exists u; apply: set1_pr; first by fprops.
    by move=> t /setI2_P [] /set2_P; case => // ->.
  exists r; apply /intersection_singletonP;fprops.
move=> x xN; move: (zprod2_pr0 xN) => [pa /set1_P <- _ _].
by ex_tac; exact: (zprod2_pr2 xN).
Qed.

Definition zmap f a b := sub f (zprod2 a b) /\ 
  (forall x, inc x (a \cup b)-> exists !z, inc z f /\ inc x z).

Definition ziequivalent a b := disjoint a b /\ exists f, zmap f a b.


Lemma zmap_symm f a b: zmap f a b -> zmap f b a.
Proof. rewrite /zmap setU2_C /zprod2 set2_C //. Qed.

Lemma zequiv_symm a b: ziequivalent a b -> ziequivalent b a.
Proof. 
move => []; rewrite /disjoint setI2_C; move=> h [f zf].
by move: (zmap_symm zf) => zf1; split => //; exists f.
Qed.

Lemma zmap_example1  a b: a <> b ->
  let A := singleton a in
    let B := singleton b in
    zmap (singleton (doubleton a b)) A B.
Proof. 
move=> nab A B; rewrite /A/B.
move: (setU2_11 a b); set (u := doubleton a b) => aux.
split.  
  by move=> t /set1_P ->; apply /zprod2_P; rewrite  aux /u;split => //;
   [exists a | exists b]; apply:set1_pr; try (apply /setI2_P;split;fprops);
     move => z /setI2_P [_] /set1_P.
rewrite aux; move=> x xu; exists u; split => //; first split; fprops.
by move => t [] /set1_P.
Qed.

Lemma zmap_example2 a b c d: let A := doubleton a b in
  let B := doubleton c d in
    disjoint A B -> a <> b -> c <> d ->
    zmap (doubleton (doubleton a c) (doubleton b d)) A B.
Proof. 
move=>  A B dab nab ncd.
move: dab; rewrite /disjoint /A/B=> ie.
have nac: a <> c by  move => ac; empty_tac1 a; rewrite ac; fprops.
have nad: a <> d by  move => ac; empty_tac1 a;rewrite ac; fprops.  
have nbc: b <> c by  move => ac; empty_tac1 b;rewrite ac; fprops.
have nbd: b <> d by  move => ac; empty_tac1 b;rewrite ac; fprops.  
have Hx: forall x y z,  y <> z ->
    singletonp (intersection2 (doubleton x y) (doubleton x z)).
  move=> x y z yz; exists x; apply: set1_pr; first by fprops.
  move => t /setI2_P []/set2_P; case => // -> /set2_P; case => //.
split.
  move=> t t1; apply / zprod2_P; case /set2_P: t1 =>  ->; split => //; 
    try (move => w /set2_P; case => ->; fprops).
   - by apply: Hx; apply: nesym.
   - by rewrite (set2_C a c); apply: Hx.
   - by rewrite (set2_C a b); apply: Hx; apply:nesym.
   - by rewrite set2_C (set2_C c d); apply: Hx.
have ->: ((doubleton a b) \cup (doubleton c d))  =
    (doubleton a c) \cup (doubleton b d). 
   set_extens t => /setU2_P; case => /set2_P; case => ->; fprops. 
move => x xu; apply /unique_existence; split.
   case /setU2_P: xu => h;
    [ exists  (doubleton a c) | exists  (doubleton b d) ]; split;fprops.
move=> u v  [] /set2_P pa qa [] /set2_P pb qb; move: qa qb.
have aux: inc x (doubleton a c) -> inc x (doubleton b d) -> False. 
  by case /set2_P => ->; case /set2_P => //; apply:nesym.
by case: pa => ->; case: pb =>-> // qc qd; case: aux.
Qed.

Lemma zequiv_example1 a b: a <> b ->
  ziequivalent (singleton a) (singleton b).
Proof. 
move=> ab; split; first by  apply: disjoint_pr => u /set1_P -> /set1_P.
exists (singleton (doubleton a b)); apply: (zmap_example1 ab).
Qed.

Lemma zequiv_example2  a b c d: let A := doubleton a b in
  let B := doubleton c d in
    disjoint A B -> a <> b -> c <> d ->
    ziequivalent A B.
Proof.
move=> A B dAB nab ncd; split => //. 
exists (doubleton (doubleton a c) (doubleton b d)); by apply: zmap_example2.
Qed.

Definition zbijective F a b :=
  [/\ (forall x, inc x a -> inc (F x) b) ,
      (forall x x', inc x a -> inc x' a -> F x = F x' -> x = x') &
      (forall y, inc y b -> exists2 x, inc x a & F x = y)].


Lemma zmap_example3 F a b: disjoint a b -> zbijective F a b ->
  ziequivalent a b.
Proof.
move=> dab [pa pb pc]; split => //. 
set (f := Zo (powerset (union2 a b)) (fun z => exists2 x, inc x a & 
    z = doubleton x (F x))).
red in dab.
exists f; split. 
  move=> t => /Zo_P  [] /setP_P tu [x xa td]; apply /zprod2_P.
  move: (pa _ xa)=> fb; split => //.
    apply: (@is_singleton_int _ _ x) => //; first by rewrite td;fprops.
    rewrite td => u;case/set2_P =>// qa qb; empty_tac1 u. 
    apply /setI2_P;split => //;ue.
  apply: (@is_singleton_int _ _  (F x)) => //; first by rewrite td;fprops.
  rewrite td => u; case /set2_P =>// qa qb; empty_tac1 u. 
  apply /setI2_P;split => //;ue.
move=> x; aw => xu; apply /unique_existence; split.
  case /setU2_P: xu => xu. 
    exists (doubleton x (F x)); split; last by fprops.
    apply: Zo_i; last by ex_tac.
    apply /setP_P;move => t;case /set2_P=> ->; [ fprops | apply /setU2_P].
    by right; apply: pa.
  move: (pc _ xu)=> [z za Fz]; exists (doubleton z x);split;fprops.
  apply: Zo_i; first by apply /setP_P;move=> t; case /set2_P=> ->; fprops.
  ex_tac; ue.
move=> u v [uf xz] [yf xy].
move: uf yf => /Zo_P [] /setP_P uu [z za zu] /Zo_P [] /setP_P vy [w wa vw].
move: xz xy;rewrite zu vw; case /set2_P => xpa; case  /set2_P=> xpb.
- by rewrite -xpa -xpb.
- by empty_tac1 x; apply /setI2_P;split => //; ue.
- by empty_tac1 x;  apply /setI2_P; split => //; ue; rewrite xpa; apply: pa.
- by rewrite xpa in xpb; rewrite (pb _ _ za wa xpb).
Qed.


Lemma zmap_pr1 f a b: zmap f a b -> union f = a \cup b.
Proof. 
move=>  [pa pb]; set_extens t.
  move /setU_P => [y ty yf]; move: (pa _ yf) => /zprod2_P.
   by move=> [yu _ _]; apply: yu.
move => tu; move: (pb _ tu) => [z [[zf xz] _]]; union_tac.
Qed.

Definition zmap_aux f x := select  (fun z => inc x z) f.

Lemma zmap_aux_pr1 f a b x:
  zmap f a b -> inc x (a \cup b) ->
  (inc (zmap_aux f x) f /\ inc x (zmap_aux f x)).
Proof.
move=> [pa pb] xu; move: (pb _ xu) => [u [[ta tb] pd]].
have h: singl_val2 (inc^~ f) (fun z => inc x z).
  by move => t y sa sb sc sd; rewrite -(pd _ (conj sa sb)) -(pd _ (conj sc sd)).
rewrite /zmap_aux - (select_uniq h ta tb);split => //.
Qed.

Lemma zmap_aux_pr2 f a b x y:
  zmap f a b -> inc x (a \cup b) -> inc y f -> inc x y ->
  y = (zmap_aux f x).
Proof. 
move =>  zm xu yf xy;move: (zmap_aux_pr1 zm xu) => [pa pb]. 
move: zm=> [_ pc]; move: (pc _ xu) => [z [zx unq]]. 
by rewrite - (unq _ (conj yf xy)) (unq _ (conj pa pb)).
Qed.

Lemma zmap_aux_pr3a f a b x y:
  zmap f a b -> inc x f -> inc y f ->  zpr x a = zpr y a  ->
  x = y.
Proof. 
move=> zm xf yf szp; move: (zm) => [pa pb].
have zxu: (inc  (zpr x a) (union2 a b)).
  by apply: setU2_1; apply: (zprod2_pr0aa (pa _ xf)).  
have zyu: (inc  (zpr y a) (union2 a b)) by ue.
rewrite (zmap_aux_pr2 zm zxu xf (zprod2_pr0ax (pa _ xf))).
by rewrite (zmap_aux_pr2 zm zyu yf (zprod2_pr0ax (pa _ yf))) szp.
Qed.

Lemma zmap_aux_pr3b f a b x y:
  zmap f a b -> inc x f -> inc y f ->  zpr x b = zpr y b ->
  x = y.
Proof. move=> zm; apply: (zmap_aux_pr3a (zmap_symm zm)). Qed.

Definition zmap_val f a x:= zpr (zmap_aux f x) a.

Lemma zmap_val_pr1a f a b x: zmap f a b ->  inc x (a \cup b) -> 
  inc (zmap_val f a x) a.
Proof. 
move=> zm xu;move: (zmap_aux_pr1 zm xu) => [pc pd]. 
move: (zm) => [pa pb]; apply: (zprod2_pr0aa (pa _ pc)).
Qed.

Lemma zmap_val_pr1b  f a b x: zmap f a b ->  inc x (a \cup b) -> 
  inc (zmap_val f b x) b.
Proof. 
move => zmf xu; move: (zmap_aux_pr1 zmf xu) => [zf xz].
move: zmf => [pa pb];move: (pa _ zf); apply: zprod2_pr0bb.
Qed.

Lemma zmap_val_pr2a  f a b x: zmap f a b ->  inc x a ->
  (zmap_val f a x) = x.
Proof. 
move => cx xa.
have  xu: (inc x (union2 a b)) by apply: setU2_1.  
move : (zmap_aux_pr1 cx xu) => [zf xz]; move: cx => [pa pb].
symmetry; apply: (zprod2_pr1a  (pa _ zf) xa xz).
Qed.

Lemma zmap_val_pr2b  f a b x: zmap f a b ->  inc x b ->
  (zmap_val f b x) = x.
Proof. 
move => cx xb.
have  xu: (inc x (union2 a b)) by apply: setU2_2.  
move : (zmap_aux_pr1 cx xu) => [zf xz]; move: cx => [pa pb].
symmetry; apply: (zprod2_pr1b  (pa _ zf) xb xz).
Qed.

Lemma zmap_val_pr3a  f a b x: zmap f a b ->  inc x a ->
  (zmap_val f a (zmap_val f b x)) = x.
Proof. 
move=> zm xa.
have xu: inc x (union2 a b) by apply: setU2_1. 
move: (zmap_aux_pr1 zm xu) => [zf xz].
move: (zm) => [pa pb]; move: (pa _ zf) => xx.
move:  (zprod2_pr1 xx) => [pc pd].
set (t := zmap_val f b x); set y := zmap_aux f x.
have: inc t (y \cap b) by rewrite pd; fprops.
move /setI2_P => [ta1 ta2].
have tab: (inc t (a \cup b)) by apply: setU2_2.
suff: y = zmap_aux f t.
   rewrite /zmap_val; move => <-;symmetry; apply: (zprod2_pr1a (pa _ zf) xa xz).
apply: (zmap_aux_pr2 zm tab zf ta1).
Qed.

Lemma zmap_val_pr3b f a b x: zmap f a b ->  inc x b ->
  (zmap_val f b (zmap_val f a x)) = x.
Proof.  
move =>  zm; have: (zmap f b a) by apply: zmap_symm.  
apply: zmap_val_pr3a.
Qed.

Lemma zmap_bijective f a b: zmap f a b -> 
  zbijective (zmap_val f b) a b.
Proof.
move=> zm; rewrite /zbijective; split => //. 
    move=> x xa; apply: (zmap_val_pr1b zm (setU2_1 b xa)).
  move=> x x' xa x'a sv;  move: (f_equal (zmap_val f a) sv). 
  rewrite zmap_val_pr3a //  zmap_val_pr3a //.
move=> y yb; exists ( zmap_val f a y).
  by apply: (zmap_val_pr1a zm); apply /setU2_2.
by apply: zmap_val_pr3b.
Qed.

Lemma zmap_setP a b: exists s,
  forall f, zmap f a b <-> inc f s.
Proof.
exists (Zo (powerset (zprod2 a b)) (fun z => zmap z a b)).
move=> f; split; last by move =>  /Zo_P [].
by  move => h; apply /Zo_P;split => //; apply /setP_P;move: h => [].
Qed.

Lemma sub_disjoint a b a' b':
  disjoint a b -> sub a' a -> sub b' b -> disjoint a' b'.
Proof. 
move=>  di sa sb; apply: disjoint_pr => u ua ub.
red in di; empty_tac1 u; aw;split => //.
Qed. 

Lemma zmap_sub f a b a': zmap f a b -> sub a' a ->
  exists f' b', [/\ sub f' f, sub b' b & zmap f' a' b'].
Proof. 
move=> zm saa.
move: (zm) => [pa pb].
set (f' := Zo f (fun z => inc (zpr z a) a')).
set (b':= Zo b (fun z => exists2 x, inc x f' & z = zpr x b)).
have sf: sub f' f by apply: Zo_S.
have sb: sub b' b by apply: Zo_S.
exists f'; exists b'; split => //. 
have pc:  (forall y, inc y f' ->  y = doubleton (zpr y a) (zpr y b)).
   by move=> y yf'; apply: zprod2_pr2; apply: pa; apply: sf.
split.
  move => x xf'.
  have ha: inc x (zprod2 a b) by apply: pa; apply: sf.
  have hb: (inc (zpr x a) a') by  move: xf' => /Zo_P [].
  have hc: (inc (zpr x b) b') by apply: Zo_i; [apply:(zprod2_pr0bb ha)| ex_tac].
  move: (zprod2_pr1 ha) => [pd pe].
  apply /zprod2_P;split => //.
      rewrite (pc _ xf');  move=> t;case /set2_P => ->; fprops.
    apply: (@is_singleton_int _ _ (zpr x a)) => //.
      apply: (@setI2_1 _ a); rewrite pd; fprops.
    move=> u ux ua'. 
    have : (inc u (x \cap a)) by fprops.
    by rewrite pd => /set1_P.
  apply: (@is_singleton_int _ _ (zpr x b)) => //.
    apply: (@setI2_1 _ b); rewrite pe; fprops.
  move=> u ux ua'. 
  have : (inc u (x \cap b)) by fprops.
  by rewrite pe => /set1_P.
move=> x xu.
have xu': (inc x (a \cup b)) by  move: xu; case /setU2_P => xs; fprops.
move: (zmap_aux_pr1 zm xu') => [pd pe]; exists (zmap_aux f x); split. 
    split => //; apply: Zo_i => //.
  move: xu; case /setU2_P => xs. 
    by move: (zmap_val_pr2a zm (saa _ xs)); rewrite /zmap_val; move => ->.
  move: xs => /Zo_P [xb [z zf' xv]].
  move: zf' => /Zo_P  [pf pg].
  have  <- //: (z = zmap_aux f x).  
  apply: (zmap_aux_pr2 zm xu' pf); rewrite xv; apply: (zprod2_pr0bx (pa _ pf)).
move=> u [uf' ux]; move: (pb _ xu') => [z [zv zu]].
by rewrite - (zu _ (conj pd pe)) (zu _ (conj (sf _ uf') ux)).
Qed.

Lemma zequiv_sub a b a': ziequivalent a b -> sub a' a ->
  exists b',  (sub b' b /\ ziequivalent a' b').
Proof.
move=>  [pa [f zm]] a'a.
move:  (zmap_sub zm a'a) => [f' [b' [pb pc pd]]]; exists b';split => //.
split; [ apply: (sub_disjoint pa a'a pc) |  by exists f'].
Qed.

Lemma zmap_transitive a b c: disjoint a c ->
  ziequivalent a b -> ziequivalent b c -> ziequivalent a c.
Proof. 
move=> dac [dab [f zmf]] [dbc [g zmg]].
move: (zmap_bijective zmf) (zmap_bijective zmg) => [pa pb pc][pd pe pf].
set (f1 := zmap_val f b);  set (f2 := zmap_val g c). 
set (h := fun x => f2 (f1 x)).
suff : (zbijective h a c) by apply: (zmap_example3 dac).
split => //; first by move=> x xa; apply: pd; apply: pa.
  fprops.
move=> y yc; move: (pf _ yc) => [x xb xb1].
by move: (pc _ xb) => [u ua uv]; ex_tac; rewrite - xb1 -uv.
Qed.

Lemma disjointness M: exists N,
  sub N M /\ ~ inc N M.
Proof. 
set (N := Zo M (fun x => ~ (inc x x))).
have NM: (sub N M) by apply: Zo_S.
exists N; split => // nm.
have aux: (~ (inc N N))by move=> bad; move: (bad) => /Zo_P [].
by apply: (aux); apply: Zo_i.
Qed.

Lemma disjointness1  M N: exists r,
  let M1 := zprod2 M (singleton r) in 
   [/\ ~ (inc r M) , disjoint M M1 & disjoint N M1].
Proof. 
move: (disjointness (M \cup (union (M \cup N)))) => [K [pa pb]].
have : (~ (inc K (union (M \cup N)))) by  dneg k; fprops.  
have kM: (~ inc K M) by dneg km; fprops.
have d1: (disjoint (union2 M N) (zprod2 M (singleton K))).
  apply: disjoint_pr => u uu us.
  move: (zprod_singleton kM) => [pc pd]; move: (pd _ us) => [v vM ud]. 
  case: pb; apply: setU2_2; apply /setU_P; exists u => //; rewrite ud; fprops.
move=> nku;exists K;split => //;apply: disjoint_pr => u um  aux; empty_tac1 u.
Qed.

Lemma zmap_example4 M r:
  let N := zprod2 M (singleton r) in 
    ~ (inc r M) -> disjoint M N ->
    ziequivalent M N.
Proof.
move=> N h1 h2. 
move: (zprod_singleton h1) => [pa pb].
suff : (zbijective (fun x => doubleton x r) M N) by apply: (zmap_example3 h2).
split => //.
  move=> x x' xM xM' sf.
  have : (inc x (doubleton x r)) by fprops.
  rewrite sf;case /set2_P => // aux;case: h1; ue.
by move=> y yN; move: (pb _ yN) => [x xM xv]; exists x.
Qed.

Lemma zequiv_example4 M N: exists M',
  disjoint N M' /\ ziequivalent M M'.
Proof. 
move: (disjointness1 M N) => [r [a pa pb]].
by exists (zprod2 M (singleton r)); split => //; apply: zmap_example4.
Qed.


Lemma zequiv_empty M: ziequivalent M emptyset -> M = emptyset.
Proof.
move=>  [_ [f zm]]; apply /set0_P => x xM.
have xu: (inc x (M \cup emptyset)) by  rewrite setU2_0.
by move: (in_set0 (zmap_val_pr1b zm xu)). 
Qed.

Lemma zequiv_no_graph M: nonempty M ->
  ~ (exists S, forall M',  ziequivalent M M' -> inc M' S).
Proof. 
move=> nM [S hs].
move: (zequiv_example4 M (union S)) => [M' [ds emm']].
move: (hs _ emm') => ms'.
case: (emptyset_dichot M') => me.
  rewrite me in emm'; move: (zequiv_empty emm') => em.
  by move: nM; rewrite em; case /nonemptyP.
move: me => [x xm]; empty_tac1 x; union_tac.  
Qed.

Definition zequiv  M N := exists2 R, ziequivalent M R & ziequivalent N R.

Lemma zequiv_reflexive M: zequiv M M.
Proof.  by move: (zequiv_example4 M M) =>[N [_ nv]]; exists N. Qed.

Lemma zequiv_symmetric M N:
   zequiv  M N -> zequiv  N M.
Proof. by move=>  [f [fa fb]]; exists f. Qed.

Lemma zequiv_transitive M N P:
   zequiv  M N -> zequiv N  P -> zequiv M P.
Proof. 
move=>  [f fa fb] [g ga gb].
move:  (zequiv_example4 f (g \cup (P \cup (M \cup N)))) => [h [hd he]].
red in hd.
exists h => //.
  have d1: (disjoint M h) by apply: disjoint_pr => u uM uh; empty_tac1 u; aw.
  exact: (zmap_transitive d1 fa he).
have d2: (disjoint N h) by apply: disjoint_pr => u uM uh; empty_tac1 u; aw. 
move: (zmap_transitive d2 fb he) => Nh.
have d3: (disjoint g h) by apply: disjoint_pr => u uM uh; empty_tac1 u; aw.
move: (zmap_transitive d3 (zequiv_symm ga) Nh) => gh.
have d4: (disjoint P h) by apply: disjoint_pr => u uM uh; empty_tac1 u; aw.
exact (zmap_transitive d4 gb gh). 
Qed.

Lemma zequiv_equipotent M N: zequiv M N <->  M \Eq N.
Proof.
have Ha: forall a b, ziequivalent a b -> a \Eq b.
  move => a b [dab [f zf]].
  exists (Lf (zmap_val f b) a b); split; aw; apply:lf_bijective.
  - move => t ta; exact:(zmap_val_pr1b zf (setU2_1 b ta)). 
  - move => u v ua va => zz.
    by rewrite - (zmap_val_pr3a zf ua)  - (zmap_val_pr3a zf va) zz.
  - move => y yb; rewrite - (zmap_val_pr3b zf yb). 
    have va:=(zmap_val_pr1a zf (setU2_2 a yb)); ex_tac.
have Hb: forall x y, disjoint x y -> x \Eq y -> ziequivalent x y.
  move => x y dxy [f [[[ff ijf] [_ sjf]] sf tf]].
  suff: zbijective (Vf f) x y by apply:zmap_example3.
  split. 
  - move => t tx; Wtac.
  - by rewrite - sf.
  - by rewrite - sf - tf.
split.
  move => [x xA xB]; exact: (EqT (Ha _ _ xA) (EqS (Ha _ _ xB))).
move => MN. move: (zequiv_example4 M N) => [M' [h1 h2]].
exists M' => //; apply:(Hb _ _ h1); exact: (EqT (EqS MN) (Ha _ _ h2)).
Qed.

End ZermeloEq.
